import sys
from contextlib import nullcontext
import numpy as np
import numexpr as ne
from numba import njit, prange
from numba.extending import overload
import dask
import dask.array as da
from dask.distributed import performance_report
from dask.diagnostics import ProgressBar
from ducc0.fft import r2c, c2r, good_size
iFs = np.fft.ifftshift
Fs = np.fft.fftshift
from daskms import xds_from_storage_ms as xds_from_ms
from daskms import xds_from_storage_table as xds_from_table
from daskms.experimental.zarr import xds_from_zarr
from omegaconf import ListConfig
from skimage.morphology import label
from scipy.optimize import fmin_l_bfgs_b
import xarray as xr
from scipy.linalg import solve_triangular
from africanus.constants import c as lightspeed
import jax.numpy as jnp
from jax import value_and_grad
import jax

JIT_OPTIONS = {
    "nogil": True,
    "cache": True
}

import pdb
class ForkedPdb(pdb.Pdb):
    """A Pdb subclass that may be used
    from a forked multiprocessing child

    """
    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            sys.stdin = open('/dev/stdin')
            pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin


def compute_context(scheduler, output_filename, boring=True):
    if scheduler == "distributed":
        return performance_report(filename=output_filename + "_dask_report.html")
    else:
        if boring:
            return nullcontext()
        else:
            return ProgressBar()


def kron_matvec(A, b):
    D = len(A)
    N = b.size
    x = b

    for d in range(D):
        Gd = A[d].shape[0]
        NGd = N // Gd
        X = np.reshape(x, (Gd, NGd))
        Z = A[d].dot(X).T
        x = Z.ravel()
    return x.reshape(b.shape)

@njit(parallel=False, cache=True, nogil=True)
def kron_matvec2(A, b):
    D = len(A)
    N = b.size
    x = b

    for d in range(D):
        Gd = A[d].shape[0]
        NGd = N//Gd
        X = np.reshape(x, (Gd, NGd))
        Z = np.zeros((Gd, NGd), dtype=A[0].dtype)
        Ad = A[d]
        for i in range(Gd):
            for j in range(Gd):
                for k in range(NGd):
                    Z[j, k] += Ad[i, j] * X[i, k]
        x[:] = Z.T.ravel()
    return x


def give_edges(p, q, nx, ny, nx_psf, ny_psf):
    nx0 = nx_psf//2
    ny0 = ny_psf//2

    # image overlap edges
    # left edge for x coordinate
    dxl = p - nx0
    xl = np.maximum(dxl, 0)

    # right edge for x coordinate
    dxu = p + nx0
    xu = np.minimum(dxu, nx)
    # left edge for y coordinate
    dyl = q - ny0
    yl = np.maximum(dyl, 0)
    # right edge for y coordinate
    dyu = q + ny0
    yu = np.minimum(dyu, ny)

    # PSF overlap edges
    xlpsf = np.maximum(nx0 - p , 0)
    xupsf = np.minimum(nx0 + nx - p, nx_psf)
    ylpsf = np.maximum(ny0 - q, 0)
    yupsf = np.minimum(ny0 + ny - q, ny_psf)

    return slice(xl, xu), slice(yl, yu), \
        slice(xlpsf, xupsf), slice(ylpsf, yupsf)


def get_padding_info(nx, ny, pfrac):
    npad_x = int(pfrac * nx)
    nfft = good_size(nx + npad_x, True)
    npad_xl = (nfft - nx) // 2
    npad_xr = nfft - nx - npad_xl

    npad_y = int(pfrac * ny)
    nfft = good_size(ny + npad_y, True)
    npad_yl = (nfft - ny) // 2
    npad_yr = nfft - ny - npad_yl
    padding = ((npad_xl, npad_xr), (npad_yl, npad_yr))
    unpad_x = slice(npad_xl, -npad_xr)
    unpad_y = slice(npad_yl, -npad_yr)
    return padding, unpad_x, unpad_y


def convolve2gaussres(image, xx, yy, gaussparf, nthreads=1, gausspari=None,
                      pfrac=0.5, norm_kernel=False):
    """
    Convolves the image to a specified resolution.

    Parameters
    ----------
    Image       - (nband, nx, ny) array to convolve
    xx/yy       - coordinates on the grid in the same units as gaussparf.
    gaussparf   - tuple containing Gaussian parameters of desired resolution
                  (emaj, emin, pa).
    gausspari   - initial resolution . By default it is assumed that the image
                  is a clean component image with no associated resolution.
                  If beampari is specified, it must be a tuple containing
                  gausspars for each imaging band in the same format.
    nthreads    - number of threads to use for the FFT's.
    pfrac       - padding used for the FFT based convolution.
                  Will pad by pfrac/2 on both sides of image
    norm_kernel -
    """
    nband, nx, ny = image.shape
    if gausspari is not None and len(gausspari) != nband:
        raise ValueError('gausspari must be on length nband')
    padding, unpad_x, unpad_y = get_padding_info(nx, ny, pfrac)
    ax = (1, 2)  # axes over which to perform fft
    lastsize = ny + np.sum(padding[-1])

    padding = ((0,0),) + padding
    image = np.pad(image, padding, mode='constant')
    imhat = r2c(iFs(image, axes=ax), axes=ax, forward=True, nthreads=nthreads,
                inorm=0)

    if len(gaussparf) == 3:  # single final resolution
        gausskern = Gaussian2D(xx, yy, gaussparf, normalise=norm_kernel)
        gausskern = np.pad(gausskern, padding[1:], mode='constant')
        gausskernhat = r2c(iFs(gausskern, axes=(0,1)), axes=(0,1),
                           forward=True, nthreads=nthreads, inorm=0)
        gausskernhat = np.broadcast_to(gausskernhat[None], imhat.shape)
    else:
        assert len(gaussparf) == nband
        gausskernhat = np.zeros_like(imhat)
        for b in range(nband):
            gausskern = Gaussian2D(xx, yy, gaussparf[b], normalise=norm_kernel)
            gausskern = np.pad(gausskern, padding[1:], mode='constant')
            r2c(iFs(gausskern, axes=(0,1)), out=gausskernhat[b],
                axes=(0,1), forward=True, nthreads=nthreads, inorm=0)


    # convolve to desired resolution
    if gausspari is None:
        imhat *= gausskernhat
    else:
        for b in range(nband):
            thiskern = Gaussian2D(xx, yy, gausspari[b], normalise=norm_kernel)
            thiskern = np.pad(thiskern, padding[1:], mode='constant')
            thiskernhat = r2c(iFs(thiskern, axes=(0,1)), axes=(0,1), forward=True,
                              nthreads=nthreads, inorm=0)

            convkernhat = np.zeros_like(thiskernhat)
            msk = np.abs(thiskernhat) > 0.0
            convkernhat[msk] = gausskernhat[b, msk]/thiskernhat[msk]

            imhat[b] *= convkernhat

    image = Fs(c2r(imhat, axes=ax, forward=False, lastsize=lastsize, inorm=2,
                   nthreads=nthreads), axes=ax)[:, unpad_x, unpad_y]

    return image


@dask.delayed
def fetch_poltype(corr_type):
    corr_type = set(tuple(corr_type))
    if corr_type.issubset(set([9, 10, 11, 12])):
        return 'linear'
    elif corr_type.issubset(set([5, 6, 7, 8])):
        return 'circular'


def construct_mappings(ms_name,
                       gain_name=None,
                       ipi=None,
                       cpi=None,
                       freq_min=-np.inf,
                       freq_max=np.inf,
                       FIELD_IDs=None,
                       DDIDs=None,
                       SCANs=None):
    '''
    Construct dictionaries containing per MS, FIELD, DDID and SCAN
    time and frequency mappings.

    Input:
    ms_name     - list of ms names
    gain_name   - list of paths to gains, must be in same order as ms_names
    ipi         - integrations (i.e. unique times) per output image.
                  Defaults to one per scan.
    cpi         - Channels per image. Defaults to one per spw.

    The chan <-> band mapping is determined by:

    freqs           - dict[MS][IDT] frequencies chunked by band
    fbin_idx        - dict[MS][IDT] freq bin starting indices
    fbin_counts     - dict[MS][IDT] freq bin counts
    band_mapping    - dict[MS][IDT] output bands in dataset
    freq_out        - array of linearly spaced output frequencies
                      over entire frequency range.
                      freq_out[band_mapping[MS][IDT]] gives the
                      model frequencies in a dataset

    where IDT is constructed as FIELD#_DDID#_SCAN#.

    Similarly, the row <-> time mapping is determined by:

    utimes          - dict[MS][IDT] unique times per output image
    tbin_idx        - dict[MS][IDT] time bin starting indices
    tbin_counts     - dict[MS][IDT] time bin counts
    time_mapping    - dict[MS][IDT] utimes per dataset

    '''

    if not isinstance(ms_name, list) and not isinstance(ms_name, ListConfig):
        ms_name = [ms_name]

    if gain_name is not None:
        if not isinstance(gain_name, list) and not isinstance(gain_name, ListConfig):
            gain_name = [gain_name]
        assert len(ms_name) == len(gain_name)

    # collect times and frequencies per ms and ds
    freqs = {}
    chan_widths = {}
    times = {}
    gains = {}
    radecs = {}
    antpos = {}
    poltype = {}
    uv_maxs = []
    idts = {}
    for ims, ms in enumerate(ms_name):
        xds = xds_from_ms(ms, chunks={"row": -1}, columns=('TIME','UVW'),
                          group_cols=['FIELD_ID', 'DATA_DESC_ID', 'SCAN_NUMBER'])

        # subtables
        ddids = xds_from_table(ms + "::DATA_DESCRIPTION")[0]
        fields = xds_from_table(ms + "::FIELD")[0]
        feeds = xds_from_table(ms + "::FEED")[0]
        spws = xds_from_table(ms + "::SPECTRAL_WINDOW")[0]
        pols = xds_from_table(ms + "::POLARIZATION")[0]
        ants = xds_from_table(ms + "::ANTENNA")[0]

        antpos[ms] = ants.POSITION.data
        poltype[ms] = fetch_poltype(pols.CORR_TYPE.data.squeeze())
        unique_feeds = np.unique(feeds.POLARIZATION_TYPE.values)

        if np.all([feed in "XxYy" for feed in unique_feeds]):
            feed_type = "linear"
        elif np.all([feed in "LlRr" for feed in unique_feeds]):
            feed_type = "circular"
        else:
            raise ValueError("Unsupported feed type/configuration.")

        idts[ms] = []
        freqs[ms] = {}
        times[ms] = {}
        radecs[ms] = {}
        chan_widths[ms] = {}
        for ds in xds:
            fid = ds.FIELD_ID
            ddid = ds.DATA_DESC_ID
            scanid = ds.SCAN_NUMBER
            if (FIELD_IDs is not None) and (fid not in FIELD_IDs):
                continue
            if (DDIDs is not None) and (ddid not in DDIDs):
                continue
            if (SCANs is not None) and (scanid not in SCANs):
                continue

            idt = f"FIELD{fid}_DDID{ddid}_SCAN{scanid}"
            idts[ms].append(idt)
            radecs[ms][idt] = fields.PHASE_DIR.data[fid].squeeze()

            freqs[ms][idt] = spws.CHAN_FREQ.data[ddid]
            chan_widths[ms][idt] = spws.CHAN_WIDTH.data[ddid]
            times[ms][idt] = da.atleast_1d(ds.TIME.data.squeeze())
            uvw = ds.UVW.data
            u_max = abs(uvw[:, 0]).max()
            v_max = abs(uvw[:, 1]).max()
            uv_maxs.append(da.maximum(u_max, v_max))

    # Early compute to get metadata
    times, freqs, radecs, chan_widths, uv_maxs, antpos, poltype = \
        dask.compute(times, freqs,radecs, chan_widths,uv_maxs,
                     antpos, poltype)

    uv_max = max(uv_maxs)
    all_freqs = []
    all_times = []
    freq_mapping = {}
    row_mapping = {}
    time_mapping = {}
    utimes = {}
    ms_chunks = {}
    for ims, ms in enumerate(ms_name):
        freq_mapping[ms] = {}
        row_mapping[ms] = {}
        time_mapping[ms] = {}
        utimes[ms] = {}
        ms_chunks[ms] = []
        gains[ms] = {}

        if gain_name is not None:
            gds = xds_from_zarr(gain_name[ims])

        for idt in idts[ms]:
            ilo = idt.find('FIELD') + 5
            ihi = idt.find('_')
            fid = int(idt[ilo:ihi])
            ilo = idt.find('DDID') + 4
            ihi = idt.rfind('_')
            ddid = int(idt[ilo:ihi])
            ilo = idt.find('SCAN') + 4
            scanid = int(idt[ilo:])
            freq = freqs[ms][idt]
            nchan_in = freq.size
            idx = (freq>=freq_min) & (freq<=freq_max)
            if not idx.any():
                continue
            idx0 = np.argmax(idx) # returns index of first True element
            try:
                # returns zero if not idx.any()
                assert idx[idx0]
            except Exception as e:
                continue
            freq = freq[idx]
            nchan = freq.size
            if cpi in [-1, 0, None]:
                cpit = nchan
                cpi = nchan_in
            else:
                cpit = np.minimum(cpi, nchan)
                cpi = np.minimum(cpi, nchan_in)
            freq_mapping[ms][idt] = {}
            tmp = np.arange(idx0, idx0 + nchan, cpit)
            freq_mapping[ms][idt]['start_indices'] = tmp
            if cpit != nchan:
                tmp2 = np.append(tmp, [idx0 + nchan])
                freq_mapping[ms][idt]['counts'] = tmp2[1:] - tmp2[0:-1]
            else:
                freq_mapping[ms][idt]['counts'] = np.array((nchan,), dtype=int)

            time = times[ms][idt]
            utime = np.unique(time)
            utimes[ms][idt] = utime
            all_times.append(utime)

            ntime = utimes[ms][idt].size
            if ipi in [0, -1, None]:
                ipit = ntime
            else:
                ipit = np.minimum(ipi, ntime)
            row_chunks, ridx, rcounts = chunkify_rows(time, ipit,
                                                      daskify_idx=False)
            # these are for applying gains
            # essentially the number of rows per unique time
            row_mapping[ms][idt] = {}
            row_mapping[ms][idt]['start_indices'] = ridx
            row_mapping[ms][idt]['counts'] = rcounts

            freq_idx0 = freq_mapping[ms][idt]['start_indices'][0]
            if freq_idx0 != 0:
                freq_chunks = (freq_idx0,) + tuple(freq_mapping[ms][idt]['counts'])
            else:
                freq_chunks = tuple(freq_mapping[ms][idt]['counts'])
            freq_idxf = np.sum(freq_chunks)
            if freq_idxf != nchan_in:
                freq_chunkf = nchan_in - freq_idxf
                freq_chunks += (freq_chunkf,)

            try:
                assert np.sum(freq_chunks) == nchan_in
            except Exception as e:
                raise RuntimeError("Something went wrong constructing the "
                                   "frequency mapping. sum(fchunks != nchan)")

            ms_chunks[ms].append({'row': row_chunks,
                                  'chan': freq_chunks})

            time_mapping[ms][idt] = {}
            tmp = np.arange(0, ntime, ipit)
            time_mapping[ms][idt]['start_indices'] = tmp
            if ipit != ntime:
                tmp2 = np.append(tmp, [ntime])
                time_mapping[ms][idt]['counts'] = tmp2[1:] - tmp2[0:-1]
            else:
                time_mapping[ms][idt]['counts'] = np.array((ntime,))

            if gain_name is not None:
                freq0 = freq[0]
                freqf = freq[-1]
                gdsf = [dsg for dsg in gds if dsg.FIELD_ID == fid]
                gdsfd = [dsg for dsg in gdsf if dsg.DATA_DESC_ID == ddid]
                # gains may have been solved over scans
                if 'SCAN_NUMBER' in gdsfd[0].attrs:
                    gdsfds = [dsg for dsg in gdsfd if dsg.SCAN_NUMBER == scanid]
                    try:
                        assert len(gdsfds) == 1
                    except Exception as e:
                        raise RuntimeError(f"Gain datasets don't align for "
                                           f"ms {ms} at FIELD_ID = {fid}, "
                                           f"DATA_DESC_ID = {ddid}, "
                                           f"SCAN_NUMBER = {scanid}. "
                                           f"len(gds) = {len(gdsfds)}")
                    try:
                        assert (gdsfds[0].gain_time == utime).all()
                    except Exception as e:
                        raise ValueError(f'Mismatch between gain and MS '
                                            f'utimes for {ms} at {idt}')

                    gains[ms][idt] = gdsfds[0]
                else:
                    try:
                        assert len(gdsfd) == 1
                    except Exception as e:
                        raise RuntimeError("Multiple gain datasets per "
                                           "FIELD and DDID but SCAN_NUMBER "
                                           "not in attributes")
                    t0 = utime[0]
                    tf = utime[-1]
                    gains[ms][idt] = gdsfd[0].sel(
                                        gain_time=slice(t0, tf))
            else:
                gains[ms][idt] = None

    return row_mapping, freq_mapping, time_mapping, \
           freqs, utimes, ms_chunks, gains, radecs, \
           chan_widths, uv_max, antpos, poltype


def restore_corrs(vis, ncorr):
    return da.blockwise(_restore_corrs, ('row', 'chan', 'corr'),
                        vis, ('row', 'chan'),
                        ncorr, None,
                        new_axes={"corr": ncorr},
                        dtype=vis.dtype)


def _restore_corrs(vis, ncorr):
    model_vis = np.zeros(vis.shape+(ncorr,), dtype=vis.dtype)
    model_vis[:, :, 0] = vis
    if model_vis.shape[-1] > 1:
        model_vis[:, :, -1] = vis
    return model_vis


def Gaussian2D(xin, yin, GaussPar=(1., 1., 0.), normalise=True, nsigma=5):
    '''
    xin         - grid of x coordinates
    yin         - grid of y coordinates
    GaussPar    - (emaj, emin, pa) with emaj/emin in units x and pa in radians.
    normalise   - normalise kernel to have volume 1
    nsigma      - compute kernel out to this many sigmas
    '''
    Smaj, Smin, PA = GaussPar
    A = np.array([[1. / Smaj ** 2, 0],
                  [0, 1. / Smin ** 2]])
    # R = np.array([[np.cos(PA), -np.sin(PA)],
    #               [np.sin(PA), np.cos(PA)]])
    # this parametrisation is equivalent to the above with
    # t = np.pi/2 - pa
    # use this for compatibility with fits
    R = np.array([[np.sin(PA), -np.cos(PA)],
                  [np.cos(PA), np.sin(PA)]])
    A = np.dot(np.dot(R.T, A), R)
    sOut = xin.shape
    # only compute the result out to 5 * emaj
    extent = (nsigma * Smaj)**2
    xflat = xin.squeeze()
    yflat = yin.squeeze()
    idx, idy = np.where(xflat**2 + yflat**2 <= extent)
    x = np.array([xflat[idx, idy].ravel(), yflat[idx, idy].ravel()])
    R = np.einsum('nb,bc,cn->n', x.T, A, x)
    # need to adjust for the fact that GaussPar corresponds to FWHM
    fwhm_conv = 2 * np.sqrt(2 * np.log(2))
    tmp = np.exp(-fwhm_conv * R)
    gausskern = np.zeros(xflat.shape, dtype=np.float64)
    gausskern[idx, idy] = tmp

    if normalise:
        gausskern /= np.sum(gausskern)
    return np.ascontiguousarray(gausskern.reshape(sOut),
                                dtype=np.float64)


@jax.jit
def psf_errorsq(x, data, xy):
    '''
    Returns sum of square error for best fit Gaussian to data.
    Note emaj must be larger than emin
    '''
    emaj, emin, pa = x
    A = jnp.array([[1. / emaj ** 2, 0],
                    [0, 1. / emin ** 2]])

    # R = jnp.array([[jnp.cos(pa), -jnp.sin(pa)],
    #                 [jnp.sin(pa), jnp.cos(pa)]])
    # this parametrisation is equivalent to the above with
    # t = np.pi/2 - pa
    # use this for compatibility with fits
    R = jnp.array([[jnp.sin(pa), -jnp.cos(pa)],
                    [jnp.cos(pa), jnp.sin(pa)]])
    B = jnp.dot(jnp.dot(R.T, A), R)
    Q = jnp.einsum('nb,bc,cn->n', xy.T, B, xy)
    # GaussPar should corresponds to FWHM
    fwhm_conv = 2 * jnp.sqrt(2 * jnp.log(2))
    model = jnp.exp(-fwhm_conv * Q)
    res = data - model
    return jnp.vdot(res, res)


def fitcleanbeam(psf: np.ndarray,
                 level: float = 0.5,
                 pixsize: float = 1.0,
                 extent: float = 5.0):
    """
    Find the Gaussian that approximates the PSF.
    First find the main lobe by identifying where PSF > level
    then fit Gaussian out to a radius of extent * max(x, y) where
    x and y are the coordinates where PSF > level.
    """
    nband, nx, ny = psf.shape

    # pixel coordinates
    x = -(nx//2) + np.arange(nx)
    y = -(ny//2) + np.arange(ny)
    xx, yy = np.meshgrid(x, y, indexing='ij')

    Gausspars = []
    for v in range(nband):
        # make sure psf is normalised
        if not psf[v].any():
            Gausspars.append([np.nan, np.nan, np.nan])
            continue
        psfv = psf[v] / psf[v].max()
        # find regions where psf is above level
        mask = np.where(psfv > level, 1.0, 0)

        # label all islands and find center
        islands = label(mask)
        ncenter = islands[nx // 2, ny // 2]

        # get extend of main lobe
        x = xx[islands == ncenter]
        y = yy[islands == ncenter]

        # initial guess for emaj and emin
        # x and y are reversed because of the parametrisation
        # of the 2D Gaussian (for fits)
        xdiff = np.maximum(y.max() - y.min(), 1)
        ydiff = np.maximum(x.max() - x.min(), 1)

        # initial guess for PA
        dx = x-np.mean(x)
        dy = y-np.mean(y)
        psftmp = psfv[islands == ncenter]
        mxx = np.mean(psftmp * dx**2)
        myy = np.mean(psftmp * dy**2)
        mxy = np.mean(psftmp * dx * dy)
        PA0 = np.pi/2 + 0.5 * np.arctan2(2 * mxy, mxx - myy)
        # ensure PA is in (0, pi)
        PA0 = np.maximum(PA0, 0.0)
        PA0 = np.minimum(PA0, np.pi)

        rsq = np.abs(x).max()**2 + np.abs(y).max()**2
        rrsq = xx**2 + yy**2
        idxs = rrsq < extent * rsq

        # select psf in fit region
        psfv = psfv[idxs]
        x = xx[idxs]
        y = yy[idxs]
        xy = np.vstack((x, y))
        if xdiff > ydiff:  # x is major axis
            emaj0 = xdiff
            emin0 = ydiff
            # PA0 = 0.0
        else:  # y is the major axis
            emaj0 = ydiff
            emin0 = xdiff
            # PA0 = np.pi/2
        dfunc = value_and_grad(psf_errorsq)
        p, f, d = fmin_l_bfgs_b(dfunc,
                                np.array((emaj0, emin0, PA0)),
                                args=(psfv, xy),
                                bounds=((0, None), (0, None), (0, np.pi)),
                                factr=1e7)
        if d['warnflag'] != 0:
            print('WARNING - warning flag raised during psf fit')

        if p[0] >= p[1]:  # major and minor axes correct
            emaj = p[0]
            emin = p[1]
            PA = p[2]
        else:  # major and minor axes have been swapped
            emaj = p[1]
            emin = p[0]
            PA = p[2] + np.pi/2
            print('WARNING - emaj/emin flipped in solver')

        Gausspars.append([emaj * pixsize, emin * pixsize, PA])

    return Gausspars


# utility to produce a model image from fitted components
def model_from_comps(comps, freq, mask, band_mapping, ref_freq, fitted):
    '''
    comps - (order/nband, ncomps) array of fitted components or model vals
            per band if no fitting was done.

    freq - (nchan) array of output frequencies

    mask - (nx, ny) bool array indicating where model is non-zero.
           comps need to be aligned with this mask i.e.
           Ix, Iy = np.where(mask) are the xy locations of non-zero pixels
           model[:, Ix, Iy] = comps if no fitting was done

    band_mapping - tuple containg the indices of bands mapping to the output.
                   In the case of a single spectral window we usually have
                   band_mapping = np.arange(nband).

    ref_freq - reference frequency used during fitting i.e. we made the
                design matrix using freq/ref_freq as coordinate

    fitted - bool indicating if any fitting actually happened.
    '''
    return da.blockwise(_model_from_comps_wrapper, ('out', 'nx', 'ny'),
                        comps, ('com', 'pix'),
                        freq, ('chan',),
                        mask, ('nx', 'ny'),
                        band_mapping, ('out',),
                        ref_freq, None,
                        fitted, None,
                        align_arrays=False,
                        dtype=comps.dtype)


def _model_from_comps_wrapper(comps, freq, mask, band_mapping, ref_freq, fitted):
    return _model_from_comps(comps[0][0], freq[0], mask, band_mapping, ref_freq, fitted)


def _model_from_comps(comps, freq, mask, band_mapping, ref_freq, fitted):
    freqo = freq[band_mapping]
    nband = freqo.size
    nx, ny = mask.shape
    model = np.zeros((nband, nx, ny), dtype=comps.dtype)
    if fitted:
        order, npix = comps.shape
        nband = freqo.size
        nx, ny = mask.shape
        model = np.zeros((nband, nx, ny), dtype=comps.dtype)
        w = (freqo / ref_freq).reshape(nband, 1)
        Xdes = np.tile(w, order) ** np.arange(0, order)
        beta_rec = Xdes.dot(comps)
        model[:, mask] = beta_rec
    else:
        model[:, mask] = comps[band_mapping]

    return model


def init_mask(mask, model, output_type, log):
    if model.ndim==3:
        _, nx, ny = model.shape
    elif model.ndim==2:
        nx, ny = model.shape
    if mask is None:
        print("No mask provided")
        mask = np.ones((nx, ny), dtype=output_type)
    elif mask.endswith('.fits'):
        try:
            from pfb.utils.fits import load_fits
            mask = load_fits(mask, dtype=output_type).squeeze()
            assert mask.shape == (nx, ny)
            print('Using provided fits mask')
        except Exception as e:
            print(f"No mask found at {mask}")
            raise e
    elif mask.lower() == 'model':
        mask = np.any(model, axis=0)
        print('Using model to construct mask')
    else:
        raise ValueError(f'Unsupported masking option {mask}')
    return mask


def dds2cubes(dds, nband, apparent=False, dual=True,
              modelname='MODEL', residname='RESIDUAL'):
    real_type = dds[0].DIRTY.dtype
    complex_type = np.result_type(real_type, np.complex64)
    nx, ny = dds[0].DIRTY.shape
    dirty = [da.zeros((nx, ny), chunks=(-1, -1),
                      dtype=real_type) for _ in range(nband)]
    model = [da.zeros((nx, ny), chunks=(-1, -1),
                      dtype=real_type) for _ in range(nband)]
    if residname in dds[0]:
        residual = [da.zeros((nx, ny), chunks=(-1, -1),
                            dtype=real_type) for _ in range(nband)]
    else:
        residual = None
    wsums = [da.zeros(1, dtype=real_type) for _ in range(nband)]
    if 'PSF' in dds[0]:
        nx_psf, ny_psf = dds[0].PSF.shape
        psf = [da.zeros((nx_psf, ny_psf), chunks=(-1, -1),
                        dtype=real_type) for _ in range(nband)]
    else:
        psf = None
    if 'PSFHAT' in dds[0]:
        nx_psf, nyo2_psf = dds[0].PSFHAT.shape
        psfhat = [da.zeros((nx_psf, nyo2_psf), chunks=(-1, -1),
                            dtype=complex_type) for _ in range(nband)]
    else:
        psfhat = None
    mean_beam = [da.zeros((nx, ny), chunks=(-1, -1),
                            dtype=real_type) for _ in range(nband)]
    if dual and 'DUAL' in dds[0]:
        nbasis, nymax, nxmax = dds[0].DUAL.shape
        dual = [da.zeros((nbasis, nymax, nxmax), chunks=(-1, -1, -1),
                            dtype=real_type) for _ in range(nband)]
    else:
        dual = None
    for ds in dds:
        b = ds.bandid
        if apparent:
            dirty[b] += ds.DIRTY.data
            if residname in ds:
                residual[b] += getattr(ds, residname).data
        else:
            dirty[b] += ds.DIRTY.data * ds.BEAM.data
            if residname in ds:
                residual[b] += getattr(ds, residname).data * ds.BEAM.data
        if 'PSF' in ds:
            psf[b] += ds.PSF.data
        if 'PSFHAT' in ds:
            psfhat[b] += ds.PSFHAT.data
        if modelname in ds:
            model[b] = getattr(ds, modelname).data
        if dual and 'DUAL' in ds:
            dual[b] = ds.DUAL.data
        mean_beam[b] += ds.BEAM.data * ds.WSUM.data[0]
        wsums[b] += ds.WSUM.data[0]
    wsums = da.stack(wsums).reshape(nband)
    wsum = wsums.sum()
    dirty = da.stack(dirty)/wsum
    model = da.stack(model)
    if residname in ds:
        residual = da.stack(residual)/wsum
    if 'PSF' in ds:
        psf = da.stack(psf)/wsum
    if 'PSFHAT' in ds:
        psfhat = da.stack(psfhat)/wsum
    if dual and 'DUAL' in ds:
        dual = da.stack(dual)
    for b in range(nband):
        if wsums[b]:
            mean_beam[b] /= wsums[b]
    mean_beam = da.stack(mean_beam)
    dirty, model, residual, psf, psfhat, mean_beam, wsums, dual = dask.compute(
                                                                dirty,
                                                                model,
                                                                residual,
                                                                psf,
                                                                psfhat,
                                                                mean_beam,
                                                                wsums,
                                                                dual)
    return (dirty, model, residual, psf, psfhat,
            mean_beam, wsums, dual)


def chunkify_rows(time, utimes_per_chunk, daskify_idx=False):
    utimes, time_bin_counts = np.unique(time, return_counts=True)
    n_time = len(utimes)
    if utimes_per_chunk == 0 or utimes_per_chunk == -1:
        utimes_per_chunk = n_time
    row_chunks = [np.sum(time_bin_counts[i:i+utimes_per_chunk])
                  for i in range(0, n_time, utimes_per_chunk)]
    time_bin_indices = np.zeros(n_time, dtype=np.int32)
    time_bin_indices[1::] = np.cumsum(time_bin_counts)[0:-1]
    time_bin_indices = time_bin_indices.astype(np.int32)
    time_bin_counts = time_bin_counts.astype(np.int32)
    if daskify_idx:
        time_bin_indices = da.from_array(time_bin_indices, chunks=utimes_per_chunk)
        time_bin_counts = da.from_array(time_bin_counts, chunks=utimes_per_chunk)
    return tuple(row_chunks), time_bin_indices, time_bin_counts


def l1reweight_func(model,
                    psiH=None,
                    outvar=None,
                    rmsfactor=None,
                    rms_comps=None,
                    alpha=4):
    '''
    The logic here is that weights should remain the same for model
    components that are rmsfactor times larger than the rms.
    High SNR values should experience relatively small thresholding
    whereas small values should be strongly thresholded
    '''
    psiH(model, outvar)
    mcomps = np.abs(np.sum(outvar, axis=0))
    # rms_comps can be per basis
    if isinstance(rms_comps, np.ndarray):
        return (1 + rmsfactor)/(1 + mcomps**alpha/rms_comps[:, None, None]**alpha)
    else:
        return (1 + rmsfactor)/(1 + mcomps**alpha/rms_comps**alpha)
    

@njit(nogil=True, cache=True)
def norm_diff(x, xp):
    return norm_diff_impl(x, xp)


def norm_diff_impl(x, xp):
    return NotImplementedError


# @overload(norm_diff_impl, jit_options={**JIT_OPTIONS, "parallel":True})
@overload(norm_diff_impl, jit_options={**JIT_OPTIONS})  # parallel reduction slower?
def nb_norm_diff_impl(x, xp):
    if x.ndim==3:
        def impl(x, xp):
            nband, nx, ny = x.shape
            num = 0.0
            den = 1e-12  # avoid div by zero
            for b in range(nband):
                for i in prange(nx):
                    for j in range(ny):
                        num += (x[b, i, j] - xp[b, i, j])**2
                        den += x[b, i, j]**2
            return np.sqrt(num/den)
    elif x.ndim==2:
        def impl(x, xp):
            nx, ny = x.shape
            num = 0.0
            den = 1e-12  # avoid div by zero
            for i in prange(nx):
                for j in range(ny):
                    num += (x[i, j] - xp[i, j])**2
                    den += x[i, j]**2
            return np.sqrt(num/den)
    else:
        raise ValueError("norm_diff is only implemented for 2D or 3D arrays")

    return impl



def remove_large_islands(x, max_island_size=100):
    islands = label(x.squeeze())
    num_islands = islands.max()
    for i in range(1,num_islands+1):
        msk = islands == i
        num_pix = np.sum(msk)
        if num_pix > max_island_size:
            x[msk] = 0.0
    return x


@njit(parallel=False, nogil=True, cache=True, inline='always')
def freqmul(A, x):
    nband, nx, ny = x.shape
    out = np.zeros_like(x)
    for k in range(nband):
        for l in range(nband):
            for i in range(nx):
                for j in range(ny):
                    out[k, i, j] += A[k, l] * x[l, i, j]
    return out


def setup_parametrisation(mode='id', minval=1e-5,
                          sigma=1.0, freq=None, lscale=1.0):
    '''
    Given a parametrisation x = f(s) return:

    func - operator that evaluates x
    finv - operator that evaluates s = f^{-1}(x)
    dfunc - operator that evaluates dx/ds at fixed x = x0
    dhfunc - operator that evaluates the adjoint of dfunc at x = x0
    '''
    nu = freq/np.mean(freq)
    nband = nu.size
    nudiffsq = (nu[:, None] - nu[None, :])**2
    K = sigma**2 * np.exp(-nudiffsq/(2*lscale**2))
    L = np.linalg.cholesky(K + 1e-10*np.eye(nband))
    LH = L.T
    if mode == 'id':
        def func(x):
            return freqmul(L, x)

        def finv(x):
            return solve_triangular(L, x, lower=True)

        def dfunc(x0, v):
            return freqmul(L, v)

        def dhfunc(x0, v):
            return freqmul(LH, v)
    elif mode == 'exp':
        def func(x):
            return np.exp(freqmul(L, x))

        def finv(x):
            tmp = solve_triangular(L, x, lower=True)
            return np.log(np.maximum(np.abs(tmp), minval))

        def dfunc(x0, v):
            return np.exp(freqmul(L, x0)) * freqmul(L, v)

        def dhfunc(x0, v):
            return freqmul(LH, v * np.exp(freqmul(L, x0)))

    else:
        raise ValueError(f"Unknown mode - {mode}")

    return func, finv, dfunc, dhfunc


def weight_from_sigma(sigma):
    weight = ne.evaluate('1.0/(s*s)',
                         local_dict={'s':sigma},
                         casting='same_kind')
    return weight


def combine_columns(x, y, dc, dc1, dc2):
    '''
    x   - dask array containing dc1
    y   - dask array containing dc2
    dc  - string that numexpr can evaluate
    dc1 - name of x
    dc2 - name of y
    '''
    ne.evaluate(dc,
                local_dict={dc1: x, dc2: y},
                out=x,
                casting='same_kind')
    return x


def set_image_size(
                uv_max,
                max_freq,
                field_of_view,
                super_resolution_factor,
                cell_size=None,
                nx=None, ny=None,
                psf_oversize=2.0):
    # max cell size
    cell_N = 1.0 / (2 * uv_max * max_freq / lightspeed)

    if cell_size is not None:
        cell_rad = cell_size * np.pi / 60 / 60 / 180
        if cell_N / cell_rad < 1:
            raise ValueError("Requested cell size too large. "
                             "Super resolution factor = ", cell_N / cell_rad)

    else:
        cell_rad = cell_N / super_resolution_factor
        cell_size = cell_rad * 60 * 60 * 180 / np.pi


    if nx is None:
        fov = field_of_view * 3600
        npix = int(fov / cell_size)
        npix = good_size(npix)
        while npix % 2:
            npix += 1
            npix = good_size(npix)
        nx = npix
        ny = npix
    else:
        nx = nx
        ny = ny if ny is not None else nx
        if nx%2 or ny%2:
            raise NotImplementedError('Only even number of pixels currently supported')
        cell_deg = np.rad2deg(cell_rad)
        fovx = nx*cell_deg
        fovy = ny*cell_deg

    nx_psf = good_size(int(psf_oversize * nx))
    while nx_psf % 2:
        nx_psf += 1
        nx_psf = good_size(nx_psf)

    ny_psf = good_size(int(psf_oversize * ny))
    while ny_psf % 2:
        ny_psf += 1
        ny_psf = good_size(ny_psf)
    nyo2 = ny_psf//2 + 1

    return nx, ny, nx_psf, ny_psf, cell_N, cell_rad

def taperf(shape, taper_width):
    height, width = shape
    array = np.ones((height, width))

    # Create 1D taper for both dimensions
    taper_x = np.ones(width)
    taper_y = np.ones(height)

    # Apply cosine taper to the edges
    taper_x[:taper_width] = 0.5 * (1 + np.cos(np.linspace(1.1*np.pi, 2*np.pi, taper_width)))
    taper_x[-taper_width:] = 0.5 * (1 + np.cos(np.linspace(0, 0.9*np.pi, taper_width)))

    taper_y[:taper_width] = 0.5 * (1 + np.cos(np.linspace(1.1*np.pi, 2*np.pi, taper_width)))
    taper_y[-taper_width:] = 0.5 * (1 + np.cos(np.linspace(0, 0.9*np.pi, taper_width)))

    return np.outer(taper_y, taper_x)


@njit(parallel=True)
def parallel_standard_normal(shape):
    rows, cols = shape
    result = np.empty(shape, dtype=np.float64)

    for i in prange(rows):
        for j in range(cols):
            result[i, j] = np.random.standard_normal()

    return result


def taperf(shape, taper_width):
    tapers1d = ()
    for npix in shape:
        taper = np.ones(npix)
        taper[:taper_width] = 0.5 * (1 + np.cos(np.linspace(1.1*np.pi, 2*np.pi, taper_width)))
        taper[-taper_width:] = 0.5 * (1 + np.cos(np.linspace(0, 0.9*np.pi, taper_width)))
        tapers1d += (taper,)
    return np.outer(*tapers1d)


@njit(nogil=True, cache=True, inline='always')
def _es_kernel(x, beta, k):
    return np.exp(beta*k*(np.sqrt((1-x)*(1+x)) - 1))


def dynamic_spectrum(time, freq, transient):
    '''
    Inputs:
        time      - time in seconds
        freq      - frequency in Hz
        transient - dict containing transient object parameters

    Outputs:
        dynamic spectrum for transient
    '''
    type = transient['type']
    I0 = transient['I0']
    alpha = transient['alpha']
    nu0 = transient['nu0']
    spectrum = I0 * (freq[None, :]/nu0)**alpha
    if type.lower()=='periodic':
        period = transient['period']
        phase0 = transient['phase0']
        dspec = spectrum * np.cos(2*np.pi*(time[:, None] - phase0)/period)
    elif type.lower()=='flare':
        sigma = transient['sigma']
        t0 = transient['t0']
        dspec = spectrum * np.exp(-0.5*((time[:, None] - t0)/sigma)**2)

    return dspec


def wplanar(uvw, threshold=1e-5):
    '''
    Checks if uv lie in a plane using PCA.

    '''
    cov_mat = np.cov(uvw, rowvar=False)
    eigenvalues = np.linalg.eigvalsh(cov_mat)
    # sorted in ascending order
    ratio = eigenvalues[0]/eigenvalues[-1]
    if ratio > threshold:
        return False
    else:
        return True
    
def fit_hyperplane(x, y):
    """Approximate a surface by a hyperplane in D dimensions

    inputs:
        x - D x N array of coordinates.
        y - N array of (possibly noisy) observations.
            Can be complex valued.

    outputs:
        theta - a vector of coefficients suct that X.T.dot(theta)
                is the hyperplane approximation of y and X is x
                with a row of ones appended as the final axis
    """
    D, N = x.shape
    y = y.squeeze()[None, :]
    z = np.vstack((x, y))
    cov = np.cov(z)
    s, V = np.linalg.eigh(cov)
    n = V[:, 0].conj()  # defines normal to the plane
    theta = np.zeros(D+1, dtype=y.dtype)
    for d in range(D+1):
        if d < D:
            theta[d] = -n[d]/n[-1]
        else:
            # we need to take the mean here because y can be noisy
            # i.e. we do not have a point exactly in the plane
            theta[d] = np.mean(n[None, 0:-1].dot(x)/n[-1] + y)
    return theta



from dataclasses import dataclass, field
import os.path
import glob
from typing import *
from scabha import configuratt
from collections import OrderedDict
from scabha.cargo import Parameter, _UNSET_DEFAULT
from omegaconf.omegaconf import OmegaConf

def EmptyDictDefault():
    return field(default_factory=lambda:OrderedDict())

schema = None

@dataclass
class _CabInputsOutputs(object):
    # inputs: Dict[str, Parameter]
    # outputs: Dict[str, Parameter]
    inputs: Dict[str, Parameter] = EmptyDictDefault()
    outputs: Dict[str, Parameter] = EmptyDictDefault()
    policies: Optional[Dict[str, Any]] = None


# load schema files
if schema is None:

    # *.yaml files under pfb.parser will be loaded automatically
    # files that should not be included must have a .yml extension
    files = glob.glob(os.path.join(os.path.dirname(__file__), "*.yaml"))

    structured = OmegaConf.structured(_CabInputsOutputs)

    tmp = configuratt.load_nested(files,
                                  structured=structured,
                                  config_class="PfbCleanCabs",
                                  use_cache=False)

    # tmp is a tuple containing the config object as the first element
    # and a set containing locations of .yaml configs for pfb workers
    schema = OmegaConf.create(tmp[0])



import numpy as np
import xarray as xr
from scipy.interpolate import RegularGridInterpolator
import sympy as sm
from sympy.utilities.lambdify import lambdify
from sympy.parsing.sympy_parser import parse_expr

specs = ['genesis', 'exodus']

def fit_image_cube(time, freq, image, wgt=None, nbasist=None, nbasisf=None,
                   method='poly', sigmasq=0):
    '''
    Fit the time and frequency axes of an image cube where

    time    - (ntime) time axis
    freq    - (nband) frequency axis
    image   - (ntime, nband, nx, ny) pixelated image
    wgt     - (ntime, nband) optional per time and frequency weights
    nbasist - number of time basis functions
    nbasisf - number of frequency basis functions
    method  - method to use for fitting (poly or Legendre)
    sigmasq - optional regularisation term to add to the Hessian
              to improve conditioning

    method:
    poly     - fit a monomials in time and frequency
    Legendre - fit a Legendre polynomial in time and frequency

    returns:
    coeffs  - fitted coefficients
    Ix, Iy  - pixel locations of non-zero pixels in the image
    expr    - a string representing the symbolic expression describing the fit
    params  - tuple of str, parameters to pass into function (excluding t and f)
    tfunc   - function which scales the time domain appropriately for method
    ffunc   - function which scales the frequency domain appropriately for method


    The fit is performed in scaled coordinates (t=time/ref_time,f=freq/ref_freq)
    '''
    ntime = time.size
    nband = freq.size
    ref_time = time[0]
    ref_freq = freq[0]
    import sympy as sm
    from sympy.abc import a, t, f

    if nbasist is None:
        nbasist = ntime
    else:
        assert nbasist <= ntime
    if nbasisf is None:
        nbasisf = nband
    else:
        assert nbasisf <= nband

    mask = np.any(image, axis=(0,1))  # over t and f axes
    Ix, Iy = np.where(mask)
    ncomps = Ix.size

    # components excluding zeros
    beta = image[:, :, Ix, Iy].reshape(ntime*nband, ncomps)
    if wgt is not None:
        wgt = wgt.reshape(ntime*nband, 1)
    else:
        wgt = np.ones((ntime*nband, 1), dtype=float)

    # nothing to fit
    if ntime==1 and nband==1:
        coeffs = beta
        expr = a
        params = (a,)
    elif method=='poly':
        wt = time/ref_time
        tfunc = t/ref_time
        Xfit = np.tile(wt[:, None], (nband, nbasist))**np.arange(nbasist)
        params = sm.symbols(f't(0:{nbasist})')
        expr = sum(co*t**i for i, co in enumerate(params))
        # the costant offset will always be included since nbasist is at least one
        if nband > 1:
            wf = freq/ref_freq
            ffunc = f/ref_freq
            Xf = np.tile(wf[:, None], (ntime, nbasisf-1))**np.arange(1, nbasisf)
            Xfit = np.hstack((Xfit, Xf))
            paramsf = sm.symbols(f'f(1:{nbasisf})')
            expr += sum(co*f**(i+1) for i, co in enumerate(paramsf))
            params += paramsf

    elif method=='Legendre':
        # scale to lie between -1,1 for stability
        if ntime > 1:
            tmax = time.max()
            tmin = time.min()
            wt = (time - (tmax + tmin)/2)
            wtmax = wt.max()
            wt /= wtmax
            # function to convert time to interp domain
            tfunc = (t - (tmax + tmin)/2)/wtmax
        else:
            wt = time
            tfunc = t
        Xt = np.zeros((ntime, nbasist), dtype=float)
        params = sm.symbols(f't(0:{nbasist})')
        if nbasist > 1:
            expr = 0
            for i in range(nbasist):
                vals = np.polynomial.Legendre.basis(i)(wt)
                Xt[:, i] = vals
                expr += sm.polys.orthopolys.legendre_poly(i, t)*params[i]
        else:
            Xt[...] = 1.0
            expr = params[0]
        Xfit = np.tile(Xt, (nband, 1))
        paramsf = sm.symbols(f'f(1:{nbasisf})')
        if nband > 1:
            Xf = np.zeros((nband, nbasisf - 1))
            fmax = freq.max()
            fmin = freq.min()
            wf = freq - (fmax + fmin)/2
            wfmax = wf.max()
            wf /= wfmax
            ffunc = (f - (fmax + fmin)/2)/wfmax
            for i in range(1, nbasisf):
                vals = np.polynomial.Legendre.basis(i)(wf)
                Xf[:, i-1] = vals
                expr += sm.polys.orthopolys.legendre_poly(i, f)*paramsf[i-1]
            Xf = np.tile(Xf, (ntime, 1))
            Xfit = np.hstack((Xfit, Xf))
            params += paramsf
    else:
        raise NotImplementedError(f"Method {method} not implemented")

    dirty_coeffs = Xfit.T.dot(wgt*beta)
    hess_coeffs = Xfit.T.dot(wgt*Xfit)
    # to improve conditioning
    if sigmasq:
        hess_coeffs += sigmasq*np.eye(hess_coeffs.shape[0])
    coeffs = np.linalg.solve(hess_coeffs, dirty_coeffs)

    return coeffs, Ix, Iy, str(expr), list(map(str,params)), str(tfunc),str(ffunc)


def fit_image_fscube(freq, image,
                     wgt=None, nbasisf=None,
                     method='Legendre', sigmasq=0):
    '''
    Fit the frequency axis of an image cube where

    freq    - (nband,) frequency axis
    image   - (nband, ncorr, nx, ny) pixelated image
    wgt     - (nband, ncorr) optional per time and frequency weights
    nbasisf - number of frequency basis functions
    method  - method to use for fitting (poly or Legendre)
    sigmasq - optional regularisation term to add to the Hessian
              to improve conditioning

    method:
    poly     - fit a monomials to frequency axis
    Legendre - fit a Legendre polynomial to frequency

    returns:
    coeffs  - (ncorr, nbasisf, ncomps) fitted coefficients
    Ix, Iy  - (ncomps,) pixel locations of non-zero pixels in the image
    expr    - a string representing the symbolic expression describing the fit
    params  - tuple of str, parameters to pass into function (excluding t and f)
    ffunc   - function which scales the frequency domain appropriately for method
    '''
    nband = freq.size
    ref_freq = freq[0]
    import sympy as sm
    from sympy.abc import f

    if nbasisf is None:
        nbasisf = nband
    else:
        assert nbasisf <= nband

    nband, ncorr, nx, ny = image.shape
    mask = np.any(image, axis=(0,1))  # over freq and corr axes
    Ix, Iy = np.where(mask)
    ncomps = Ix.size

    # components excluding zeros
    beta = image[:, :, Ix, Iy].reshape(nband, ncorr, ncomps)
    if wgt is not None:
        wgt = wgt.reshape(nband, ncorr, 1)
    else:
        wgt = np.ones((nband, ncorr, 1), dtype=float)

    params = sm.symbols(f'f(0:{nbasisf})')
    if nband==1:  # nothing to fit
        coeffs = beta
        expr = f
        params = (f,)
    elif method=='poly':
        wf = freq/ref_freq
        ffunc = f/ref_freq
        Xf = np.tile(wf[:, None], (1, nbasisf))**np.arange(nbasisf)
        expr = sum(co*f**i for i, co in enumerate(params))

    elif method=='Legendre':
        Xf = np.zeros((nband, nbasisf), dtype=float)
        fmax = freq.max()
        fmin = freq.min()
        wf = freq - (fmax + fmin)/2
        wfmax = wf.max()
        wf /= wfmax
        ffunc = (f - (fmax + fmin)/2)/wfmax
        Xf[:, 0] = 1.0
        expr = params[0]
        for i in range(1, nbasisf):
            vals = np.polynomial.Legendre.basis(i)(wf)
            Xf[:, i] = vals
            expr += sm.polys.orthopolys.legendre_poly(i, f)*params[i]
    else:
        raise NotImplementedError(f"Method {method} not implemented")

    # fit each correlation separately
    coeffs = np.zeros((ncorr, nbasisf, ncomps), dtype=beta.dtype)
    for c in range(ncorr):
        dirty_coeffs = Xf.T.dot(wgt[:, c]*beta[:, c])
        hess_coeffs = Xf.T.dot(wgt[:, c]*Xf)
        # to improve conditioning
        if sigmasq:
            hess_coeffs += sigmasq*np.eye(hess_coeffs.shape[0])
        coeffs[c] = np.linalg.solve(hess_coeffs, dirty_coeffs)


    return coeffs, Ix, Iy, str(expr), list(map(str,params)), str(ffunc)


def eval_coeffs_to_cube(time, freq, nx, ny, coeffs, Ix, Iy,
                        expr, paramf, texpr, fexpr):
    ntime = time.size
    nfreq = freq.size

    image = np.zeros((ntime, nfreq, nx, ny), dtype=float)
    params = sm.symbols(('t','f'))
    params += sm.symbols(tuple(paramf))
    symexpr = parse_expr(expr)
    modelf = lambdify(params, symexpr)
    texpr = parse_expr(texpr)
    tfunc = lambdify(params[0], texpr)
    fexpr = parse_expr(fexpr)
    ffunc = lambdify(params[1], fexpr)
    for i, tval in enumerate(time):
        for j, fval in enumerate(freq):
            image[i, j, Ix, Iy] = modelf(tfunc(tval), ffunc(fval), *coeffs)

    return image


def eval_coeffs_to_slice(time, freq, coeffs, Ix, Iy,
                         expr, paramf, texpr, fexpr,
                         nxi, nyi, cellxi, cellyi, x0i, y0i,
                         nxo, nyo, cellxo, cellyo, x0o, y0o):

    image_in = np.zeros((nxi, nyi), dtype=float)
    params = sm.symbols(('t','f'))
    params += sm.symbols(tuple(paramf))
    symexpr = parse_expr(expr)
    modelf = lambdify(params, symexpr)
    texpr = parse_expr(texpr)
    tfunc = lambdify(params[0], texpr)
    fexpr = parse_expr(fexpr)
    ffunc = lambdify(params[1], fexpr)
    image_in[Ix, Iy] = modelf(tfunc(time), ffunc(freq), *coeffs)

    pix_area_in = cellxi * cellyi
    pix_area_out = cellxo * cellyo
    area_ratio = pix_area_out/pix_area_in

    xin = (-(nxi//2) + np.arange(nxi))*cellxi + x0i
    yin = (-(nyi//2) + np.arange(nyi))*cellyi + y0i
    xo = (-(nxo//2) + np.arange(nxo))*cellxo + x0o
    yo = (-(nyo//2) + np.arange(nyo))*cellyo + y0o

    # how many pixels to pad by to extrapolate with zeros
    xldiff = xin.min() - xo.min()
    if xldiff > 0.0:
        npadxl = int(np.ceil(xldiff/cellxi))
    else:
        npadxl = 0
    yldiff = yin.min() - yo.min()
    if yldiff > 0.0:
        npadyl = int(np.ceil(yldiff/cellyi))
    else:
        npadyl = 0

    xudiff = xo.max() - xin.max()
    if xudiff > 0.0:
        npadxu = int(np.ceil(xudiff/cellxi))
    else:
        npadxu = 0
    yudiff = yo.max() - yin.max()
    if yudiff > 0.0:
        npadyu = int(np.ceil(yudiff/cellyi))
    else:
        npadyu = 0

    do_pad = npadxl > 0
    do_pad |= npadxu > 0
    do_pad |= npadyl > 0
    do_pad |= npadyu > 0
    if do_pad:
        image_in = np.pad(image_in,
                        ((npadxl, npadxu), (npadyl, npadyu)),
                        mode='constant')

        xin = (-(nxi//2+npadxl) + np.arange(nxi + npadxl + npadxu))*cellxi + x0i
        nxi = nxi + npadxl + npadxu
        yin = (-(nyi//2+npadyl) + np.arange(nyi + npadyl + npadyu))*cellyi + y0i
        nyi = nyi + npadyl + npadyu

    do_interp = cellxi != cellxo
    do_interp |= cellyi != cellyo
    do_interp |= x0i != x0o
    do_interp |= y0i != y0o
    do_interp |= nxi != nxo
    do_interp |= nyi != nyo
    if do_interp:
        interpo = RegularGridInterpolator((xin, yin), image_in,
                                          bounds_error=True, method='linear')
        xx, yy = np.meshgrid(xo, yo, indexing='ij')
        return interpo((xx, yy)) * area_ratio
    else:
        return image_in


def model_from_mds(mds_name):
    '''
    Evaluate component model at the original resolution
    '''
    mds = xr.open_zarr(mds_name, chunks=None)
    return eval_coeffs_to_cube(mds.times.values,
                               mds.freqs.values,
                               mds.npix_x, mds.npix_y,
                               mds.coefficients.values,
                               mds.location_x.values, mds.location_y.values,
                               mds.parametrisation,
                               mds.params.values,
                               mds.texpr, mds.fexpr)


# flake8: noqa
import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('GRID')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.grid)
def grid(**kw):
    '''
    Compute imaging weights and create a dirty image, psf etc.
    By default only the MFS images are converted to fits files.
    Set the --fits-cubes flag to also produce fits cubes.

    '''
    opts = OmegaConf.create(kw)
    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        if opts.nworkers > 1:
            ntpw = nthreads//opts.nworkers
            opts.nthreads = ntpw//2
            ncpu = ntpw//2
        else:
            opts.nthreads = nthreads//2
            ncpu = ncpu//2

    from daskms.fsspec_store import DaskMSStore
    basename = opts.output_filename
    fits_oname = f'{opts.fits_output_folder}/{oname}'
    dds_store = DaskMSStore(f'{basename}_{opts.suffix}.dds')

    if opts.xds is not None:
        xds_store = DaskMSStore(opts.xds.rstrip('/'))
        xds_name = opts.xds
    else:
        xds_store = DaskMSStore(f'{basename}.xds')
        xds_name = f'{basename}.xds'
    try:
        assert xds_store.exists()
    except Exception as e:
        log.error_and_raise(f"There must be an xds at {xds_name}. ",
                            RuntimeError)
    opts.xds = xds_store.url
    OmegaConf.set_struct(opts, True)

    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/grid_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool, thread_pool_size
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    from pfb import set_client
    if opts.nworkers > 1:
        client = set_client(opts.nworkers, log, client_log_level=opts.log_level)
        from distributed import as_completed
    else:
        log.info("Faking client")
        from pfb.utils.dist import fake_client
        client = fake_client()
        names = [0]
        as_completed = lambda x: x

    from pfb.utils.naming import xds_from_url
    from pfb.utils.fits import dds2fits

    ti = time.time()
    psfpars_mfs = _grid(**opts)

    dds, dds_list = xds_from_url(dds_store.url)

    # convert to fits files
    futures = []
    if opts.fits_mfs or opts.fits_cubes:
        log.info(f"Writing fits files to {fits_oname}_{opts.suffix}")
        if opts.dirty:
            fut = client.submit(dds2fits,
                                dds_list,
                                'DIRTY',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=True,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes,
                                psfpars_mfs=psfpars_mfs)
            futures.append(fut)
        if opts.psf:
            fut = client.submit(dds2fits,
                                dds_list,
                                'PSF',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=True,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes,
                                psfpars_mfs=psfpars_mfs)
            futures.append(fut)
        if opts.residual and 'RESIDUAL' in dds[0]:
            fut = client.submit(dds2fits,
                                dds_list,
                                'MODEL',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=False,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes,
                                psfpars_mfs=psfpars_mfs)
            futures.append(fut)
        if 'MODEL' in dds[0]:
            fut = client.submit(dds2fits,
                                dds_list,
                                'RESIDUAL',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=True,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes,
                                psfpars_mfs=psfpars_mfs)
            futures.append(fut)
        if opts.noise:
            fut = client.submit(dds2fits,
                                dds_list,
                                'NOISE',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=True,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes,
                                psfpars_mfs=psfpars_mfs)
            futures.append(fut)

        if 'BEAM' in dds[0]:
            fut = client.submit(dds2fits,
                                dds_list,
                                'BEAM',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=False,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes,
                                psfpars_mfs=psfpars_mfs)
            futures.append(fut)

        for fut in as_completed(futures):
            try:
                column = fut.result()
            except:
                continue
            log.info(f'Done writing {column}')

        log.info(f"All done after {time.time() - ti}s")

    if opts.nworkers > 1:
        try:
            client.close()
        except Exception as e:
            pass

def _grid(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import numpy as np
    from distributed import as_completed, get_client
    from itertools import cycle
    import fsspec
    from daskms.fsspec_store import DaskMSStore
    from pfb.utils.misc import set_image_size, fitcleanbeam
    from pfb.operators.gridder import image_data_products, wgridder_conventions
    import xarray as xr
    from pfb.utils.astrometry import get_coordinates
    from africanus.coordinates import radec_to_lm
    from pfb.utils.naming import xds_from_url, cache_opts, get_opts

    try:
        client = get_client()
        names = list(client.scheduler_info()['workers'].keys())
    except:
        from pfb.utils.dist import fake_client
        client = fake_client()
        names = [0]
        as_completed = lambda x: x

    basename = opts.output_filename

    # xds contains vis products, no imaging weights applied
    xds_name = f'{basename}.xds' if opts.xds is None else opts.xds
    xds_store = DaskMSStore(xds_name)
    try:
        assert xds_store.exists()
    except Exception as e:
        log.error_and_raise(f"There must be a dataset at {xds_store.url}",
                            RuntimeError)

    log.info(f"Lazy loading xds from {xds_store.url}")
    xds, xds_list = xds_from_url(xds_store.url)
    valid_bands = np.unique([ds.bandid for ds in xds])

    times_in = []
    freqs_in = []
    for ds in xds:
        times_in.append(ds.time_out)
        freqs_in.append(ds.freq_out)

    times_in = np.unique(times_in)
    freqs_out = np.unique(freqs_in)

    ntime_in = times_in.size
    nband = freqs_out.size

    real_type = xds[0].WEIGHT.dtype
    if real_type == np.float32:
        precision = 'single'
    else:
        precision = 'double'

    # max uv coords over all datasets
    uv_max = 0
    max_freq = 0
    for ds in xds:
        uv_max = np.maximum(uv_max, ds.uv_max)
        max_freq = np.maximum(max_freq, ds.max_freq)

    nx, ny, nx_psf, ny_psf, cell_N, cell_rad = set_image_size(
        uv_max,
        max_freq,
        opts.field_of_view,
        opts.super_resolution_factor,
        opts.cell_size,
        opts.nx,
        opts.ny,
        opts.psf_oversize
    )
    cell_deg = np.rad2deg(cell_rad)
    cell_size = cell_deg * 3600
    log.info(f"Super resolution factor = {cell_N/cell_rad}")
    log.info(f"Cell size set to {cell_size:.5e} arcseconds")
    log.info(f"Field of view is ({nx*cell_deg:.3e},{ny*cell_deg:.3e}) degrees")

    # create dds and cache
    dds_name = opts.output_filename + f'_{opts.suffix}' + '.dds'
    dds_store = DaskMSStore(dds_name)
    if '://' in dds_store.url:
        protocol = xds_store.url.split('://')[0]
    else:
        protocol = 'file'

    if dds_store.exists() and opts.overwrite:
        log.info(f"Removing {dds_store.url}")
        dds_store.rm(recursive=True)

    fs = fsspec.filesystem(protocol)
    if dds_store.exists() and not opts.overwrite:
        # get opts from previous run
        optsp = get_opts(dds_store.url,
                         protocol,
                         name='opts.pkl')

        # check if we need to remake the data products
        verify_attrs = ['epsilon',
                        'do_wgridding', 'double_accum',
                        'field_of_view', 'super_resolution_factor']
        try:
            for attr in verify_attrs:
                assert optsp[attr] == opts[attr]

            from_cache = True
            log.info("Initialising from cached data products")
        except Exception as e:
            log.info(f'Cache verification failed on {attr}. '
                  'Will remake image data products')
            dds_store.rm(recursive=True)
            fs.makedirs(dds_store.url, exist_ok=True)
            # dump opts to validate cache on rerun
            cache_opts(opts,
                       dds_store.url,
                       protocol,
                       name='opts.pkl')
            from_cache = False
    else:
        fs.makedirs(dds_store.url, exist_ok=True)
        cache_opts(opts,
                   dds_store.url,
                   protocol,
                   name='opts.pkl')
        from_cache = False
        log.info("Initialising from scratch.")

    log.info(f"Data products will be cached in {dds_store.url}")

    # filter datasets by time and band
    xds_dct = {}
    if opts.concat_row:
        ntime = 1
        times_out = np.mean(times_in, keepdims=True)
        for b, bid in enumerate(valid_bands):
            for ds, ds_name in zip(xds, xds_list):
                if ds.bandid == bid:
                    tbid = f'time0000_band{b:04d}'
                    xds_dct.setdefault(tbid, {})
                    xds_dct[tbid].setdefault('dsl', [])
                    xds_dct[tbid]['dsl'].append(ds_name)
                    xds_dct[tbid]['radec'] = (ds.ra, ds.dec)
                    xds_dct[tbid]['time_out'] = times_out[0]
                    xds_dct[tbid]['freq_out'] = freqs_out[b]
                    xds_dct[tbid]['chan_low'] = ds.chan_low
                    xds_dct[tbid]['chan_high'] = ds.chan_high
    else:
        ntime = ntime_in
        times_out = times_in
        for t in range(times_in.size):
            for b, bid in enumerate(valid_bands):
                for ds, ds_name in zip(xds, xds_list):
                    if ds.time_out == times_in[t] and ds.bandid == bid:
                        tbid = f'time{t:04d}_band{b:04d}'
                        xds_dct.setdefault(tbid, {})
                        xds_dct[tbid].setdefault('dsl', [])
                        xds_dct[tbid]['dsl'].append(ds_name)
                        xds_dct[tbid]['radec'] = (ds.ra, ds.dec)
                        xds_dct[tbid]['time_out'] = times_out[t]
                        xds_dct[tbid]['freq_out'] = freqs_out[b]
                        xds_dct[tbid]['chan_low'] = ds.chan_low
                        xds_dct[tbid]['chan_high'] = ds.chan_high


    ncorr = ds.corr.size
    corrs = ds.corr.values
    if opts.dirty:
        log.info(f"Image size = (ntime={ntime}, nband={nband}, "
              f"ncorr={ncorr}, nx={nx}, ny={ny})")

    if opts.psf:
        log.info(f"PSF size = (ntime={ntime}, nband={nband}, "
              f"ncorr={ncorr}, nx={nx_psf}, ny={ny_psf})")

    # check if model exists
    if opts.transfer_model_from:
        try:
            mds = xr.open_zarr(opts.transfer_model_from, chunks=None)
        except Exception as e:
            log.error_and_raise(f"No dataset found at {opts.transfer_model_from}",
                                RuntimeError)

        # should we load these inside the worker calls?
        model_coeffs = mds.coefficients.values
        locx = mds.location_x.values
        locy = mds.location_y.values
        params = mds.params.values

        log.info(f"Loading model from {opts.transfer_model_from}. ")

    futures = []
    for wname, (tbid, ds_dct) in zip(cycle(names), xds_dct.items()):
        bandid = tbid[-4:]
        timeid = tbid[4:8]
        ra = ds_dct['radec'][0]
        dec = ds_dct['radec'][1]
        dsl = ds_dct['dsl']
        time_out = ds_dct['time_out']
        freq_out = ds_dct['freq_out']
        chan_low = ds_dct['chan_low']
        chan_high = ds_dct['chan_high']
        iter0 = 0
        if from_cache:
            out_ds_name = f'{dds_store.url}/time{timeid}_band{bandid}.zarr'
            out_ds = xr.open_zarr(out_ds_name,
                                  chunks=None)
            if 'niters' in out_ds:
                iter0 = out_ds.niters
        else:
            out_ds_name = None

        # compute lm coordinates of target
        if opts.target is not None:
            tmp = opts.target.split(',')
            if len(tmp) == 1 and tmp[0] == opts.target:
                tra, tdec = get_coordinates(time_out, target=opts.target)
            else:  # we assume a HH:MM:SS,DD:MM:SS format has been passed in
                from astropy import units as u
                from astropy.coordinates import SkyCoord
                c = SkyCoord(tmp[0], tmp[1], frame='fk5', unit=(u.hourangle, u.deg))
                tra = np.deg2rad(c.ra.value)
                tdec = np.deg2rad(c.dec.value)

            tcoords=np.zeros((1,2))
            tcoords[0,0] = tra
            tcoords[0,1] = tdec
            coords0 = np.array((ra, dec))
            lm0 = radec_to_lm(tcoords, coords0).squeeze()
            l0 = lm0[0]
            m0 = lm0[1]
        else:
            l0 = 0.0
            m0 = 0.0
            tra = ds.ra
            tdec = ds.dec

        attrs = {
            'ra': tra,
            'dec': tdec,
            'l0': l0,
            'm0': m0,
            'cell_rad': cell_rad,
            'bandid': bandid,
            'timeid': timeid,
            'freq_out': freq_out,
            'time_out': time_out,
            'robustness': opts.robustness,
            'super_resolution_factor': opts.super_resolution_factor,
            'field_of_view': opts.field_of_view,
            'product': opts.product,
            'niters': iter0,
            'chan_low': chan_low,
            'chan_high': chan_high,
        }

        # get the model
        if opts.transfer_model_from:
            from pfb.utils.modelspec import eval_coeffs_to_slice
            _, _, _, x0, y0 = wgridder_conventions(l0, m0)
            model = eval_coeffs_to_slice(
                time_out,
                freq_out,
                model_coeffs,
                locx, locy,
                mds.parametrisation,
                params,
                mds.texpr,
                mds.fexpr,
                mds.npix_x, mds.npix_y,
                mds.cell_rad_x, mds.cell_rad_y,
                mds.center_x, mds.center_y,
                nx, ny,
                cell_rad, cell_rad,
                x0, y0
            )
            model = model[None, :, :]  # hack to get the corr axis

        elif from_cache:
            if opts.use_best_model and 'BEST_MODEL' in out_ds:
                model = out_ds.MODEL_BEST.values
            elif 'MODEL' in out_ds:
                model = out_ds.MODEL.values
            else:
                model = None
        else:
            model = None

        fut = client.submit(image_data_products,
                            dsl,
                            out_ds_name,
                            nx, ny,
                            nx_psf, ny_psf,
                            cell_rad, cell_rad,
                            dds_store.url + '/' + tbid + '.zarr',
                            attrs,
                            model=model,
                            robustness=opts.robustness,
                            l0=l0, m0=m0,
                            nthreads=opts.nthreads,
                            epsilon=opts.epsilon,
                            do_wgridding=opts.do_wgridding,
                            double_accum=opts.double_accum,
                            l2_reweight_dof=opts.l2_reweight_dof,
                            do_dirty=opts.dirty,
                            do_psf=opts.psf,
                            do_weight=opts.weight,
                            do_residual=opts.residual,
                            do_noise=opts.noise,
                            do_beam=opts.beam,
                            workers=wname)
        futures.append(fut)

    residual_mfs = {}
    wsum = {}
    if opts.psf:
        psf_mfs = {}
    nds = len(futures)
    n_launched = 1
    for fut in as_completed(futures):
        print(f"\rProcessing: {n_launched}/{nds}", end='', flush=True)
        outputs = fut.result()
        timeid = outputs['timeid']
        residual_mfs.setdefault(timeid, np.zeros((ncorr, nx, ny), dtype=float))
        residual_mfs[timeid] += outputs['residual']
        wsum.setdefault(timeid, np.zeros(ncorr, dtype=float))
        wsum[timeid] += outputs['wsum']
        if opts.psf:
            psf_mfs.setdefault(timeid, np.zeros((ncorr, nx_psf, ny_psf), dtype=float))
            psf_mfs[timeid] += outputs['psf']
        n_launched += 1

    print("\n")  # after progressbar above

    for timeid in residual_mfs.keys():
        # get MFS PSFPARSN
        # resolution in pixels
        if opts.psf:
            psfparsn = {}
            psf_mfs[timeid] /= wsum[timeid][:, None, None]
            psfparsn[timeid] = fitcleanbeam(psf_mfs[timeid])
        else:
            psfparsn = None

        residual_mfs[timeid] /= wsum[timeid][:, None, None]
        for c in range(ncorr):
            rms = np.std(residual_mfs[timeid][c])
            if np.isnan(rms):
                log.error_and_raise('RMS of residual in nan, something went wrong',
                                    RuntimeError)
            rmax = np.abs(residual_mfs[timeid][c]).max()
            log.info(f"Time ID {timeid}: {corrs[c]} - resid max = {rmax:.3e}, "
                  f"rms = {rms:.3e}")

    # put these in the dds for future reference
    if psfparsn is not None:
        cache_opts(psfparsn,
                   dds_store.url,
                   protocol,
                   name='psfparsn_mfs.pkl')

    return psfparsn



import numpy as np
import dask.array as da
from uuid import uuid4
from ducc0.fft import r2c, c2r


def psf_convolve_slice(
                    xpad,  # preallocated array to store padded image
                    xhat,  # preallocated array to store FTd image
                    xout,  # preallocated array to store output image
                    psfhat,
                    lastsize,
                    x,  # input image, not overwritten
                    nthreads=1):
    nx, ny = x.shape
    xpad[...] = 0.0
    xpad[0:nx, 0:ny] = x
    r2c(xpad, axes=(0, 1), nthreads=nthreads,
        forward=True, inorm=0, out=xhat)
    xhat *= psfhat
    c2r(xhat, axes=(0, 1), forward=False, out=xpad,
        lastsize=lastsize, inorm=2, nthreads=nthreads,
        allow_overwriting_input=True)
    xout[...] = xpad[0:nx, 0:ny]
    return xout


def psf_convolve_cube(xpad,    # preallocated array to store padded image
                      xhat,    # preallocated array to store FTd image
                      xout,    # preallocated array to store output image
                      psfhat,
                      lastsize,
                      x,       # input image, not overwritten
                      nthreads=1):
    '''
    The copyto is not necessarily faster it just allows us to see where time is spent
    '''
    _, nx, ny = x.shape
    xpad[...] = 0.0
    # np.copyto(xpad, 0.0)
    xpad[:, 0:nx, 0:ny] = x
    # np.copyto(xpad[:, 0:nx, 0:ny], x)
    r2c(xpad, axes=(1, 2), nthreads=nthreads,
        forward=True, inorm=0, out=xhat)
    xhat *= psfhat
    # ne.evaluate('xhat*psfhat', out=xhat, casting='unsafe')
    c2r(xhat, axes=(1, 2), forward=False, out=xpad,
        lastsize=lastsize, inorm=2, nthreads=nthreads,
        allow_overwriting_input=True)
    xout[...] = xpad[:, 0:nx, 0:ny]
    # np.copyto(xout, xpad[:, 0:nx, 0:ny])
    return xout


def psf_convolve_fscube(xpad,    # preallocated array to store padded image
                        xhat,    # preallocated array to store FTd image
                        xout,    # preallocated array to store output image
                        psfhat,
                        lastsize,
                        x,       # input image, not overwritten
                        nthreads=1):
    _, ncorr, nx, ny = x.shape
    xpad[...] = 0.0
    xpad[:, :, 0:nx, 0:ny] = x
    r2c(xpad, axes=(-2, -1), nthreads=nthreads,
        forward=True, inorm=0, out=xhat)
    xhat *= psfhat
    c2r(xhat, axes=(-2, -1), forward=False, out=xpad,
        lastsize=lastsize, inorm=2, nthreads=nthreads,
        allow_overwriting_input=True)
    xout[...] = xpad[:, :, 0:nx, 0:ny]
    return xout



def psf_convolve_xds(x, xds, psfopts, wsum, eta, mask,
                     compute=True, use_beam=True):
    '''
    Image space Hessian with reduction over dataset
    '''
    if not isinstance(x, da.Array):
        x = da.from_array(x, chunks=(1, -1, -1), name=False)

    if not isinstance(mask, da.Array):
        mask = da.from_array(mask, chunks=(-1, -1), name=False)

    assert mask.ndim==2

    nband, nx, ny = x.shape

    convims = [da.zeros((nx, ny),
               chunks=(-1, -1), name="zeros-" + uuid4().hex)
               for _ in range(nband)]

    for ds in xds:
        psfhat = ds.PSFHAT.data
        b = ds.bandid
        if use_beam:
            beam = ds.BEAM.data * mask
        else:
            # TODO - separate implementation without
            # unnecessary beam application
            beam = mask

        convim = psf_convolve(x[b], psfhat, beam, psfopts)

        convims[b] += convim

    convim = da.stack(convims)/wsum

    if eta:
        convim += x * eta**2

    if compute:
        return convim.compute()
    else:
        return convim




##################### jax #####################################
import jax
import jax.numpy as jnp
from functools import partial

@partial(jax.jit, static_argnums=(0,1,2,3))
def psf_convolve_slice_jax(
                    nx, ny,
                    nx_psf, ny_psf,
                    psfhat,
                    x):
    psfh = jax.lax.stop_gradient(psfhat)
    xhat = jnp.fft.rfft2(x,
                         s=(nx_psf, ny_psf),
                         norm='backward')
    xout = jnp.fft.irfft2(xhat*psfh,
                          s=(nx_psf, ny_psf),
                          norm='backward')[0:nx, 0:ny]
    return xout



import numpy as np


class Mask(object):
    def __init__(self, mask):
        """
        Mask operator
        """
        nx, ny = mask.shape
        self.nx = nx
        self.ny = ny
        self.mask = mask

    def dot(self, x):
        """
        Components to image
        """
        return x[~self.mask]

    def hdot(self, x):
        """
        Image to components
        """
        res = np.zeros((self.nx, self.ny))
        res[~self.mask] = x
        return res



import numpy as np
import numexpr as ne
import xarray as xr
from numba import literally
from dask.graph_manipulation import clone
from distributed import worker_client
from operator import getitem
from pfb.utils.beam import interp_beam
from pfb.utils.weighting import weight_data
from uuid import uuid4
import gc
from casacore.quanta import quantity
from datetime import datetime, timezone
from katbeam import JimBeam
from scipy import ndimage
from scipy.constants import c as lightspeed


def single_stokes(
                dc1=None,
                dc2=None,
                operator=None,
                ds=None,
                jones=None,
                opts=None,
                freq=None,
                chan_width=None,
                utime=None,
                tbin_idx=None,
                tbin_counts=None,
                chan_low=None,
                chan_high=None,
                radec=None,
                antpos=None,
                poltype=None,
                xds_store=None,
                bandid=None,
                timeid=None,
                msid=None,
                wid=None):

    fieldid = ds.FIELD_ID
    ddid = ds.DATA_DESC_ID
    scanid = ds.SCAN_NUMBER
    oname = f'ms{msid:04d}_fid{fieldid:04d}_spw{ddid:04d}_scan{scanid:04d}' \
            f'_band{bandid:04d}_time{timeid:04d}'
    
    if opts.precision.lower() == 'single':
        real_type = np.float32
        complex_type = np.complex64
    elif opts.precision.lower() == 'double':
        real_type = np.float64
        complex_type = np.complex128

    with worker_client() as client:
        (ds, jones) = client.compute([clone(ds),
                                      clone(jones)],
                                     sync=True,
                                     workers=wid,
                                     key='read-'+uuid4().hex)
    data = getattr(ds, dc1).values
    ds = ds.drop_vars(dc1)
    if dc2 is not None:
        try:
            assert (operator=='+' or operator=='-')
        except Exception as e:
            raise e
        ne.evaluate(f'data {operator} data2',
                    local_dict={'data': data,
                                'data2': getattr(ds, dc2).values},
                    out=data,
                    casting='same_kind')
        ds = ds.drop_vars(dc2)

    time = ds.TIME.values
    ds = ds.drop_vars('TIME')
    interval = ds.INTERVAL.values
    ds = ds.drop_vars('INTERVAL')
    ant1 = ds.ANTENNA1.values
    ds = ds.drop_vars('ANTENNA1')
    ant2 = ds.ANTENNA2.values
    ds = ds.drop_vars('ANTENNA2')
    uvw = ds.UVW.values
    ds = ds.drop_vars('UVW')
    flag = ds.FLAG.values
    ds = ds.drop_vars('FLAG')
    # MS may contain auto-correlations
    frow = ds.FLAG_ROW.values | (ant1 == ant2)
    ds = ds.drop_vars('FLAG_ROW')

    # combine flag and frow
    flag = np.logical_or(flag, frow[:, None, None])

    # we rely on this to check the number of output bands and
    # to ensure we don't end up with fully flagged chunks
    if flag.all():
        return None

    nrow, nchan, ncorr = data.shape

    if opts.sigma_column is not None:
        weight = ne.evaluate('1.0/sigma**2',
                             local_dict={'sigma': getattr(ds, opts.sigma_column).values})
        ds = ds.drop_vars(opts.sigma_column)
    elif opts.weight_column is not None:
        weight = getattr(ds, opts.weight_column).values
        ds = ds.drop_vars(opts.weight_column)
    else:
        weight = np.ones((nrow, nchan, ncorr),
                         dtype=real_type)

    # this seems to help with memory consumption
    # note the ds.drop_vars above
    del ds
    gc.collect()

    nrow, nchan, ncorr = data.shape
    ntime = utime.size
    nant = antpos.shape[0]
    time_out = np.mean(utime)

    freq_out = np.mean(freq)
    freq_min = freq.min()
    freq_max = freq.max()

    if data.dtype != complex_type:
        data = data.astype(complex_type)

    if weight.dtype != real_type:
        weight = weight.astype(real_type)

    if jones is not None:
        if jones.dtype != complex_type:
            jones = jones.astype(complex_type)
        # qcal has chan and ant axes reversed compared to pfb implementation
        jones = np.swapaxes(jones, 1, 2)
        # data are not 2x2 so we need separate labels
        # for jones correlations and data/weight correlations
        # reshape to dispatch with overload
        jones_ncorr = jones.shape[-1]
        if jones_ncorr == 4:
            jones = jones.reshape(ntime, nant, nchan, 1, 2, 2)
        elif jones_ncorr == 2:
            pass
        else:
            raise ValueError("Incorrect number of correlations of "
                            f"{jones_ncorr} for product {opts.product}")
    else:
        jones = np.ones((ntime, nant, nchan, 1, 2),
                        dtype=complex_type)

    # check that there are no missing antennas
    ant1u = np.unique(ant1)
    ant2u = np.unique(ant2)
    allants = np.unique(np.concatenate((ant1u, ant2u)))

    # check that antpos gives the correct size table
    antmax = allants.size
    if opts.check_ants:
        try:
            assert antmax == nant
        except Exception as e:
            raise ValueError('Inconsistent ANTENNA table. '
                            'Shape does not match max number of antennas '
                            'as inferred from ant1 and ant2. '
                            f'Table size is {antpos.shape} but got {antmax}. '
                            f'{oname}')

    # relabel antennas by index
    # this only works because allants is sorted in ascending order
    for a, ant in enumerate(allants):
        ant1 = np.where(ant1==ant, a, ant1)
        ant2 = np.where(ant2==ant, a, ant2)

    # apply gains and convert to Stokes
    data, weight = weight_data(data, weight, flag, jones,
                            tbin_idx, tbin_counts,
                            ant1, ant2,
                            literally(poltype),
                            literally(opts.product),
                            literally(str(ncorr)))
    
    # TODO - check if wsum for any of the correlations is zero
    # This happens e.g. if selecting out diagonal correlations
    # with QC and making CORRECTED_WEIGHTS 

    # do after weight_data otherwise mappings need to be recomputed
    # drop fully flagged rows
    mrow = ~frow
    data = data[mrow]
    time = time[mrow]
    interval = interval[mrow]
    ant1 = ant1[mrow]
    ant2 = ant2[mrow]
    uvw = uvw[mrow]
    flag = flag[mrow]
    weight = weight[mrow]
    gc.collect()

    # number of output correlations will be set by required Stokes products
    ncorr = data.shape[-1]
    # we need this for averaging
    flag = np.tile(flag.any(axis=-1, keepdims=True), (1,1,ncorr))

    # do before averaging
    uv_max = np.maximum(np.abs(uvw[:, 0]).max(), np.abs(uvw[:, 1]).max())
    max_freq = freq.max()

    # set corr coords (removing duplicates and sorting)
    corr = list("".join(dict.fromkeys(sorted(opts.product))))
    ncorr = len(corr)
    
    # simple average over channels
    if opts.chan_average > 1:
        from africanus.averaging import time_and_channel

        res = time_and_channel(
                    time,
                    interval,
                    ant1,
                    ant2,
                    uvw=uvw,
                    flag=flag,
                    weight_spectrum=weight,
                    visibilities=data,
                    chan_freq=freq,
                    chan_width=chan_width,
                    time_bin_secs=1e-15,
                    chan_bin_size=opts.chan_average)

        data = res.visibilities
        weight = res.weight_spectrum
        flag = res.flag
        freq = res.chan_freq
        chan_width = res.chan_width
        uvw = res.uvw
        nchan = freq.size

    if opts.bda_decorr < 1:
        from africanus.averaging import bda
        res = bda(time,
                  interval,
                  ant1, ant2,
                  uvw=uvw,
                  flag=flag,
                  weight_spectrum=weight,
                  visibilities=data,
                  chan_freq=freq,
                  chan_width=chan_width,
                  decorrelation=opts.bda_decorr,
                  min_nchan=freq.size,
                  max_fov=opts.max_field_of_view)

        offsets = res.offsets
        uvw = res.uvw[offsets[:-1], :]
        weight = res.weight_spectrum.reshape(-1, nchan, ncorr)
        data = res.visibilities.reshape(-1, nchan, ncorr)
        flag = res.flag.reshape(-1, nchan, ncorr)

    flag = flag.any(axis=-1)
    mask = (~flag).astype(np.uint8)


    # TODO - better beam interpolation
    fov = opts.max_field_of_view
    cell_rad = 1.0 / (uv_max * max_freq / lightspeed)
    cell_deg = np.rad2deg(cell_rad)
    npix = int(fov/cell_deg)
    l_beam = (-(npix//2) + np.arange(npix)) * cell_deg
    m_beam = (-(npix//2) + np.arange(npix)) * cell_deg
    if opts.beam_model is None:
        beam = np.ones((ncorr, npix, npix), dtype=real_type)
    elif opts.beam_model.lower() == 'katbeam':
        if freq_min >= 8.5e8 and freq_max <= 1.8e9:
            beamo = JimBeam('MKAT-AA-L-JIM-2020')
        elif freq_min >= 5.4e8 and freq_max <= 1.1e9:
            beamo = JimBeam('MKAT-AA-UHF-JIM-2020')
        # elif freq_min >= 8.56e8 and freq_max <= 1.71179102e+09:
        #     beamo = JimBeam('MKAT-AA-S-JIM-2020')
        else:
            raise ValueError(f"Freq range not covered by katbeam")
        xx, yy = np.meshgrid(l_beam, m_beam, indexing='ij')
        # katbeam expects freq in MHz
        fMHz = freq_out/1e6
        beam = np.zeros((ncorr, npix, npix), dtype=np.float64)
        for i, product in enumerate(corr):
            beam0 = getattr(beamo, product)(xx, yy, fMHz)
            step = 25
            angles = np.linspace(0, 359, step)
            for angle in angles:
                beam[i] += ndimage.rotate(beam0, angle, reshape=False, mode='nearest')
            beam[i] /= angles.size
            # how to normalise the center for other Stokes products?
            # beam[i] /= beam[i].max()
    else:
        raise ValueError(f"Unknown beam model {opts.beam_model}")

    # for operations that follow it will be preferable to have the corr axis
    # first for contiguity
    data_vars = {}
    data_vars['VIS'] = (('corr', 'row', 'chan'), data.transpose(2, 0, 1))
    data_vars['WEIGHT'] = (('corr', 'row', 'chan'), weight.transpose(2, 0, 1))
    data_vars['MASK'] = (('row', 'chan'), mask)
    data_vars['UVW'] = (('row', 'three'), uvw)
    data_vars['FREQ'] = (('chan',), freq)
    data_vars['BEAM'] = (('corr', 'l_beam','m_beam'), beam)

    coords = {'chan': (('chan',), freq),
              'l_beam': (('l_beam',), l_beam),
              'm_beam': (('m_beam',), m_beam),
              'corr': (('corr',), corr)
    }

    unix_time = quantity(f'{time_out}s').to_unix_time()
    utc = datetime.fromtimestamp(unix_time, tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')

    attrs = {
        'ra' : radec[0],
        'dec': radec[1],
        'fieldid': fieldid,
        'ddid': ddid,
        'scanid': scanid,
        'freq_out': freq_out,
        'freq_min': freq_min,
        'freq_max': freq_max,
        'chan_low': chan_low,
        'chan_high': chan_high,
        'bandid': bandid,
        'time_out': time_out,
        'time_min': utime.min(),
        'time_max': utime.max(),
        'timeid': timeid,
        'product': opts.product,
        'utc': utc,
        'max_freq': max_freq,
        'uv_max': uv_max,
        'beam_model': opts.beam_model
    }

    out_ds = xr.Dataset(data_vars, coords=coords,
                        attrs=attrs)
    out_ds.to_zarr(f'{xds_store}/{oname}.zarr',
                                mode='w')
    return time_out, freq_out



import numpy as np
import numexpr as ne
from pfb.utils.misc import give_edges
from pfb.utils import logging as pfb_logging
log = pfb_logging.get_logger('HOGBOM')


def hogbom(
        ID,
        PSF,
        threshold=0,
        gamma=0.1,
        pf=0.1,
        maxit=10000,
        report_freq=1000,
        verbosity=1):
    nband, nx, ny = ID.shape
    _, nx_psf, ny_psf = PSF.shape
    nx0 = nx_psf//2
    ny0 = ny_psf//2
    x = np.zeros((nband, nx, ny), dtype=ID.dtype)
    IR = ID.copy()
    IRsearch = np.sum(IR, axis=0)**2
    pq = IRsearch.argmax()
    p = pq//ny
    q = pq - p*ny
    IRmax = np.sqrt(IRsearch[p, q])
    wsums = np.amax(PSF, axis=(1,2))
    fsel = wsums > 0
    tol = np.maximum(pf * IRmax, threshold)
    k = 0
    stall_count = 0
    while IRmax > tol and k < maxit and stall_count < 5:
        xhat = IR[fsel, p, q] / wsums[fsel]
        x[:, p, q] += gamma * xhat
        ne.evaluate('IR - gamma * xhat * psf', local_dict={
                    'IR': IR,
                    'gamma': gamma,
                    'xhat': xhat[:, None, None],
                    'psf': PSF[:, nx0 - p:nx0 + nx - p,
                                  ny0 - q:ny0 + ny - q]},
                    out=IR, casting='same_kind')
        IRsearch = np.sum(IR, axis=0)**2
        pq = IRsearch.argmax()
        p = pq//ny
        q = pq - p*ny
        IRmaxp = IRmax
        IRmax = np.sqrt(IRsearch[p, q])
        k += 1

        if np.abs(IRmaxp - IRmax) / np.abs(IRmaxp) < 5e-3:
            stall_count += stall_count

        if not k % report_freq and verbosity > 1:
            log.info("At iteration %i max residual = %f" % (k, IRmax))

    IRmfs = np.sum(IR, axis=0)
    rms = np.std(IRmfs[~np.any(x, axis=0)])

    if k >= maxit:
        if verbosity:
            log.info(f"Max iters reached. "
                  f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return x, 1
    elif stall_count >= 5:
        if verbosity:
            log.info(f"Stalled. "
                  f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return x, 1
    else:
        if verbosity:
            log.info(f"Success, converged after {k} iterations. "
                  f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return x, 0


# import jax.numpy as jnp
# from jax import jit
# from jax.ops import index_add
# import jax.lax as lax
# @jit
# def hogbom_jax(ID, PSF, x, gamma=0.1, pf=0.1, maxit=5000):
#     nx, ny = ID.shape
#     IR = jnp.array(ID, copy=True)
#     IRsearch = jnp.square(IR)
#     pq = jnp.argmax(IRsearch)
#     p = pq//ny
#     q = pq - p*ny
#     IRmax = jnp.sqrt(IRsearch[p, q])
#     tol = pf*IRmax
#     k = 0

#     def cond_func(inputs):

#         IRmax, IR, IRsearch, PSF, x, loc, tol, gamma, k = inputs

#         return (k < maxit) & (IRmax > tol)

#     def body_func(inputs):
#         IRmax, IR, IRsearch, PSF, x, loc, tol, gamma, k = inputs
#         nx, ny = IR.shape
#         p, q = loc
#         xhat = IR[p, q]
#         x = index_add(x, (p, q), gamma * xhat)
#         modconv = lax.dynamic_slice(PSF, [nx-p, ny-q], [nx, ny])
#         IR = IR - gamma * xhat * modconv
#         IRsearch = jnp.square(IR)
#         pq = IRsearch.argmax()
#         p = pq//ny
#         q = pq - p*ny
#         IRmax = jnp.sqrt(IRsearch[p, q])
#         return (IRmax, IR, IRsearch, PSF, x, (p, q), tol, gamma, k+1)

#     init_val = (IRmax, IR, IRsearch, PSF, x, (p, q), tol, gamma, k)
#     out = lax.while_loop(cond_func, body_func, init_val)

#     return out[4], out[1]



# flake8: noqa
import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
import time
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema
import ray
import fsspec

log = pfb_logging.get_logger('HCI')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.hci)
def hci(**kw):
    '''
    Produce high cadence residual images.
    '''
    opts = OmegaConf.create(kw)

    output_dataset = opts.output_dataset

    if '://' in output_dataset:
        protocol = output_dataset.split('://')[0]
        prefix = f'{protocol}://'
    else:
        protocol = 'file'
        prefix = ''

    fs = fsspec.filesystem(protocol)
    basedir = fs.expand_path('/'.join(output_dataset.split('/')[:-1]))[0]
    if not fs.exists(basedir):
        fs.makedirs(basedir)

    oname = output_dataset.split('/')[-1]

    opts.output_dataset = f'{prefix}{basedir}/{oname}'

    # this should be a file system
    opts.log_directory = basedir

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        opts.nthreads = nthreads//2
        ncpu = ncpu//2

    remprod = opts.product.upper().strip('IQUV')
    if len(remprod):
        log.error_and_raise(f"Product {remprod} not yet supported",
                            NotImplementedError)

    from daskms.fsspec_store import DaskMSStore
    msnames = []
    for ms in opts.ms:
        msstore = DaskMSStore(ms.rstrip('/'))
        mslist = msstore.fs.glob(ms.rstrip('/'))
        try:
            assert len(mslist) > 0
            msnames += list(map(msstore.fs.unstrip_protocol, mslist))
        except:
            log.error_and_raise(f"No MS at {ms}",
                                ValueError)
    opts.ms = msnames
    if opts.gain_table is not None:
        gainnames = []
        for gt in opts.gain_table:
            gainstore = DaskMSStore(gt.rstrip('/'))
            gtlist = gainstore.fs.glob(gt.rstrip('/'))
            try:
                assert len(gtlist) > 0
                gainnames += list(map(gainstore.fs.unstrip_protocol, gtlist))
            except Exception as e:
                log.error_and_raise(f"No gain table  at {gt}",
                                    ValueError)
        opts.gain_table = gainnames

    OmegaConf.set_struct(opts, True)

    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/hci_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    if opts.object_store_memory is not None:
        mem_limit = opts.object_store_memory * 1e9  # convert GB -> B
    else:
        mem_limit = None

    ray.init(num_cpus=opts.nworkers,
             logging_level='INFO',
             ignore_reinit_error=True,
             object_store_memory=mem_limit,
             local_mode=opts.nworkers==1)

    ti = time.time()
    _hci(**opts)

    log.info(f"All done after {time.time() - ti}s")

    ray.shutdown()

def _hci(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import numpy as np
    from pfb.utils.misc import construct_mappings
    from daskms import xds_from_storage_ms as xds_from_ms
    import fsspec
    from daskms.fsspec_store import DaskMSStore
    import dask.array as da
    from africanus.constants import c as lightspeed
    from ducc0.fft import good_size
    from pfb.utils.stokes2im import batch_stokes_image
    import xarray as xr
    import yaml
    from tempfile import TemporaryDirectory
    from zarr import ProcessSynchronizer

    basename = f'{opts.output_dataset}'

    if opts.stack and opts.output_format == 'fits':
        raise RuntimeError("Can't stack in fits mode")

    fds_store = DaskMSStore(basename)
    if fds_store.exists():
        if opts.overwrite:
            log.info(f"Overwriting {basename}")
            fds_store.rm(recursive=True)
        else:
            log.error_and_raise(f"{basename} exists. "
                                "Set overwrite to overwrite it. ",
                                RuntimeError)

    fs = fsspec.filesystem(fds_store.url.split(':', 1)[0])
    fs.makedirs(fds_store.url, exist_ok=True)

    if opts.gain_table is not None:
        tmpf = lambda x: '::'.join(x.rsplit('/', 1))
        gain_names = list(map(tmpf, opts.gain_table))
    else:
        gain_names = None

    if opts.freq_range is not None:
        fmin, fmax = opts.freq_range.strip(' ').split(':')
        if len(fmin) > 0:
            freq_min = float(fmin)
        else:
            freq_min = -np.inf
        if len(fmax) > 0:
            freq_max = float(fmax)
        else:
            freq_max = np.inf
    else:
        freq_min = -np.inf
        freq_max = np.inf

    group_by = ['FIELD_ID', 'DATA_DESC_ID', 'SCAN_NUMBER']

    # write model to tmp ds
    if opts.transfer_model_from is not None:
        log.error_and_raise('Use the degrid app to populate a model column instead',
                            NotImplementedError)
        
    ipc = opts.integrations_per_chunk
    if ipc % opts.integrations_per_image:
        log.warning("Warning - integrations-per-image does not divide integrations-per-chunk evenly")

    log.info('Constructing mapping')
    row_mapping, freq_mapping, time_mapping, \
        freqs, utimes, ms_chunks, gains, radecs, \
        chan_widths, uv_max, antpos, poltype = \
            construct_mappings(opts.ms,
                               gain_names,
                               ipi=ipc,
                               cpi=opts.channels_per_image,
                               freq_min=freq_min,
                               freq_max=freq_max,
                               FIELD_IDs=opts.fields,
                               DDIDs=opts.ddids,
                               SCANs=opts.scans)

    max_freq = 0

    # get the full (time, freq) domain
    all_times = []
    all_freqs = []
    for ms in opts.ms:
        for idt in freqs[ms].keys():
            freq = freqs[ms][idt]
            all_freqs.extend(freq)
            all_times.extend(utimes[ms][idt])
            mask  = (freq <= freq_max) & (freq >= freq_min)
            max_freq = np.maximum(max_freq, freq[mask].max())

    all_freqs = np.unique(all_freqs)
    all_times = np.unique(all_times)

    # cell size
    cell_N = 1.0 / (2 * uv_max * max_freq / lightspeed)

    if opts.cell_size is not None:
        cell_size = opts.cell_size
        cell_rad = cell_size * np.pi / 60 / 60 / 180
        if cell_N / cell_rad < 1:
            log.info(f"Requested cell size of {cell_size} arcseconds could be sub-Nyquist.")
        log.info(f"Super resolution factor = {cell_N/cell_rad}")
    else:
        cell_rad = cell_N / opts.super_resolution_factor
        cell_size = cell_rad * 60 * 60 * 180 / np.pi
        log.info(f"Cell size set to {cell_size} arcseconds")

    if opts.nx is None:
        fov = opts.field_of_view * 3600
        npix = int(fov / cell_size)
        npix = good_size(npix)
        while npix % 2:
            npix += 1
            npix = good_size(npix)
        nx = npix
        ny = npix
    else:
        nx = opts.nx
        ny = opts.ny if opts.ny is not None else nx
        cell_deg = np.rad2deg(cell_rad)
        if nx%2 or ny%2:
            raise NotImplementedError('Only even number of pixels currently supported')
        fovx = nx*cell_deg
        fovy = ny*cell_deg
        log.info(f"Field of view is ({fovx:.3e},{fovy:.3e}) degrees")

    cell_deg = np.rad2deg(cell_rad)

    # crude column arithmetic
    dc = opts.data_column.replace(" ", "")
    if "+" in dc:
        dc1, dc2 = dc.split("+")
        operator="+"
    elif "-" in dc:
        dc1, dc2 = dc.split("-")
        operator="-"
    else:
        dc1 = dc
        dc2 = None
        operator=None

    columns = (dc1,
               opts.flag_column,
               'UVW', 'ANTENNA1',
               'ANTENNA2', 'TIME', 'INTERVAL', 'FLAG_ROW')
    schema = {}
    schema[opts.flag_column] = {'dims': ('chan', 'corr')}
    schema[dc1] = {'dims': ('chan', 'corr')}
    if dc2 is not None:
        columns += (dc2,)
        schema[dc2] = {'dims': ('chan', 'corr')}

    # only WEIGHT column gets special treatment
    # any other column must have channel axis
    if opts.sigma_column is not None:
        log.info(f"Initialising weights from {opts.sigma_column} column")
        columns += (opts.sigma_column,)
        schema[opts.sigma_column] = {'dims': ('chan', 'corr')}
    elif opts.weight_column is not None:
        log.info(f"Using weights from {opts.weight_column} column")
        columns += (opts.weight_column,)
        # hack for https://github.com/ratt-ru/dask-ms/issues/268
        if opts.weight_column != 'WEIGHT':
            schema[opts.weight_column] = {'dims': ('chan', 'corr')}
    else:
        log.info(f"No weights provided, using unity weights")

    # distinct freq groups
    sgroup = 0
    freq_groups = []
    freq_sgroups = []
    for ms in opts.ms:
        for idt, freq in freqs[ms].items():
            ilo = idt.find('DDID') + 4
            ihi = idt.rfind('_')
            ddid = int(idt[ilo:ihi])
            if (opts.ddids is not None) and (ddid not in opts.ddids):
                continue
            if not len(freq_groups):
                freq_groups.append(freq)
                freq_sgroups.append(sgroup)
                sgroup += freq_mapping[ms][idt]['counts'].size
            else:
                in_group = False
                for fs in freq_groups:
                    if freq.size == fs.size and np.all(freq == fs):
                        in_group = True
                        break
                if not in_group:
                    freq_groups.append(freq)
                    freq_sgroups.append(sgroup)
                    sgroup += freq_mapping[ms][idt]['counts'].size

    # band mapping
    msddid2bid = {}
    for ms in opts.ms:
        msddid2bid[ms] = {}
        for idt, freq in freqs[ms].items():
            # find group where it matches
            for sgroup, fs in zip(freq_sgroups, freq_groups):
                if freq.size == fs.size and np.all(freq == fs):
                    msddid2bid[ms][idt] = sgroup

    # construct examplar dataset if asked to stack
    if opts.stack and opts.output_format != 'zarr':
        raise ValueError('Can only stack zarr outputs not fits')
    attrs, ntasks, n_timeo, n_freqo = make_dummy_dataset(
            opts, fds_store.url, utimes, freqs, radecs,
            time_mapping, freq_mapping,
            freq_min, freq_max, nx, ny, cell_deg, ipc)

    n_stokes = len(opts.product)
    log.info(f"Cube size = (n_stokes={n_stokes}, n_freq={n_freqo}, n_time={n_timeo}, n_y={ny}, n_x={nx})")

    if opts.inject_transients is not None:
        # we need to do this here because we don't have access
        # to the full domain inside the call to stokes2im
        log.info("Computing transient spectra")
        from pfb.utils.transients import generate_transient_spectra
        with open(opts.inject_transients, 'r') as f:
            config = yaml.safe_load(f)
        transients = config['transients']
        coords = {
            "TIME": (("TIME",), all_times),
            "FREQ": (("FREQ",), all_freqs),
        }
        names = []
        ras = []
        decs = []
        transient_ds = xr.Dataset(coords=coords)
        for transient in transients:
            tprofile, fprofile = generate_transient_spectra(all_times, all_freqs, transient)
            name = transient["name"]
            transient_ds[f'{name}_time_profile'] = (("TIME",), tprofile)
            transient_ds[f'{name}_freq_profile'] = (("FREQ",), fprofile)
            names.append(name)
            ras.append(transient["position"]["ra"])
            decs.append(transient["position"]["dec"])
        transient_ds = transient_ds.assign_attrs({"names": names, "ras": ras, "decs": decs})
        transient_baseoname = opts.inject_transients.removesuffix('yaml')
        transient_ds.to_zarr(f"{transient_baseoname}zarr", mode='a')
        log.info("Spectra computed")

    if opts.model_column is not None:
        columns += (opts.model_column,)
        schema[opts.model_column] = {'dims': ('chan', 'corr')}

    tasks = []
    channel_width = {}
    ncompleted = 0
    # if opts.temp-dir is None dir will be /tmp
    with TemporaryDirectory(prefix='hci-', dir=opts.temp_dir) as tempdir:
        synchronizer = ProcessSynchronizer(tempdir)
        for ims, ms in enumerate(opts.ms):
            xds = xds_from_ms(ms,
                            columns=columns,
                            table_schema=schema,
                            group_cols=group_by)

            for ds in xds:
                fid = ds.FIELD_ID
                ddid = ds.DATA_DESC_ID
                scanid = ds.SCAN_NUMBER
                if (opts.fields is not None) and (fid not in opts.fields):
                    continue
                if (opts.ddids is not None) and (ddid not in opts.ddids):
                    continue
                if (opts.scans is not None) and (scanid not in opts.scans):
                    continue

                idt = f"FIELD{fid}_DDID{ddid}_SCAN{scanid}"

                idx = (freqs[ms][idt]>=freq_min) & (freqs[ms][idt]<=freq_max)
                if not idx.any():
                    continue

                titr = enumerate(zip(time_mapping[ms][idt]['start_indices'],
                                    time_mapping[ms][idt]['counts']))
                for ti, (tlow, tcounts) in titr:

                    It = slice(tlow, tlow + tcounts)
                    ridx = row_mapping[ms][idt]['start_indices'][It]
                    rcnts = row_mapping[ms][idt]['counts'][It]
                    # select all rows for output dataset
                    Irow = slice(ridx[0], ridx[-1] + rcnts[-1])

                    fitr = enumerate(zip(freq_mapping[ms][idt]['start_indices'],
                                            freq_mapping[ms][idt]['counts']))
                    b0 = msddid2bid[ms][idt]
                    for fi, (flow, fcounts) in fitr:
                        Inu = slice(flow, flow + fcounts)

                        subds = ds[{'row': Irow, 'chan': Inu}]
                        subds = subds.chunk({'row':-1, 'chan': -1})
                        if gains[ms][idt] is not None:
                            subgds = gains[ms][idt][{'gain_time': It, 'gain_freq': Inu}]
                            jones = subgds.gains.data
                        else:
                            jones = None
                        
                        fut = batch_stokes_image.remote(
                                dc1=dc1,
                                dc2=dc2,
                                operator=operator,
                                ds=subds,
                                jones=jones,
                                opts=opts,
                                nx=nx,
                                ny=ny,
                                freq=freqs[ms][idt][Inu],
                                utime=utimes[ms][idt][It],
                                rbin_idx=ridx,
                                rbin_counts=rcnts,
                                cell_rad=cell_rad,
                                radec=radecs[ms][idt],
                                antpos=antpos[ms],
                                poltype=poltype[ms],
                                fds_store=fds_store,
                                bandid=b0+fi,
                                timeid=ti,
                                msid=ims,
                                attrs=attrs,
                                integrations_per_image=opts.integrations_per_image,
                                synchronizer=synchronizer
                        )
                        tasks.append(fut)
                        channel_width[b0+fi] = freqs[ms][idt][Inu].max() - freqs[ms][idt][Inu].min()

                        # limit the number of chunks that are processed simultaneously
                        if len(tasks) > opts.max_simul_chunks:
                            ncompleted += 1
                            ready, tasks = ray.wait(tasks, num_returns=1)
                            timeid, bandid = ray.get(ready)[0]
                            print(f"Processed {ncompleted}/{ntasks}", end='\n', flush=True)

        remaining_tasks = tasks.copy()
        while remaining_tasks:
            ncompleted += 1
            ready, remaining_tasks = ray.wait(remaining_tasks, num_returns=1)
            timeid, bandid = ray.get(ready)[0]
            print(f"Processed {ncompleted}/{ntasks}", end='\n', flush=True)

    if opts.stack:
        log.info("Computing means")
        import dask
        import dask.array as da
        from concurrent.futures import ThreadPoolExecutor
        cwidths = []
        for _, val in channel_width.items():
            cwidths.append(val)
        # reduction over FREQ and TIME so use max chunk sizes
        ds = xr.open_zarr(fds_store.url, chunks={'FREQ': -1, 'TIME':-1})
        cube = ds.cube.data
        wsums = ds.weight.data
        # all variables have been normalised by wsum so we first
        # undo the normalisation (wsum=0 where there is no data)
        weighted_cube = cube * wsums[:, :, :, None, None]
        taxis = ds.cube.get_axis_num('TIME')
        faxis = ds.cube.get_axis_num('FREQ')
        wsum = da.sum(wsums, axis=taxis)
        # we need this for the where clause in da.divide, should be cheap
        wsumc = wsum.compute()[:, :, None, None]
        weighted_sum = da.sum(weighted_cube, axis=taxis)
        weighted_mean = da.divide(weighted_sum, wsum[:, :, None, None], where=wsumc>0)
        if opts.psf_out:
            psfsq = (ds.psf.data*wsums[:, :, :, None, None])**2
            weighted_psfsq_sum = da.sum(psfsq, axis=(faxis, taxis))
            wsumsq = da.sum(wsums**2, axis=(faxis, taxis))
            wsumsqc = wsumsq.compute()[:, None, None]
            weighted_psfsq_mean = da.divide(weighted_psfsq_sum, wsumsq[:, None, None], where=wsumsqc>0)
            if opts.psf_relative_size == 1:
                xpsf = "X"
                ypsf = "Y"
            else:
                xpsf = "X_PSF"
                ypsf = "Y_PSF"
            ds['psf2'] = (('STOKES', ypsf, xpsf),  weighted_psfsq_mean)
        # only write new variables
        drop_vars = [key for key in ds.data_vars.keys() if key != 'psf2']
        ds = ds.drop_vars(drop_vars)
        ds['cube_mean'] = (('STOKES', 'FREQ', 'Y', 'X'),  weighted_mean)
        ds['channel_width'] = (('FREQ',), da.from_array(cwidths, chunks=1))
        with dask.config.set(pool=ThreadPoolExecutor(8)):
            ds.to_zarr(fds_store.url, mode='r+')
        log.info("Reduction complete")
    return

def make_dummy_dataset(opts, cds_url, utimes, freqs, radecs, time_mapping, freq_mapping,
                       freq_min, freq_max, nx, ny, cell_deg, time_chunk,
                       spatial_chunk=128):
    import numpy as np
    import dask.array as da
    import xarray as xr
    from daskms import xds_from_storage_ms as xds_from_ms
    from ducc0.fft import good_size
    out_ra = []
    out_dec = []
    out_times = []
    out_freqs = []
    ntasks = 0
    for ms in opts.ms:
        xds = xds_from_ms(ms,
                    columns='TIME',
                    group_cols=['FIELD_ID', 'DATA_DESC_ID', 'SCAN_NUMBER'])
        for ds in xds:
            fid = ds.FIELD_ID
            ddid = ds.DATA_DESC_ID
            scanid = ds.SCAN_NUMBER
            if (opts.fields is not None) and (fid not in opts.fields):
                continue
            if (opts.ddids is not None) and (ddid not in opts.ddids):
                continue
            if (opts.scans is not None) and (scanid not in opts.scans):
                continue

            idt = f"FIELD{fid}_DDID{ddid}_SCAN{scanid}"

            idx = (freqs[ms][idt]>=freq_min) & (freqs[ms][idt]<=freq_max)
            if not idx.any():
                continue

            out_ra.append(radecs[ms][idt][0])
            out_dec.append(radecs[ms][idt][1])

            titr = enumerate(zip(time_mapping[ms][idt]['start_indices'],
                                time_mapping[ms][idt]['counts']))
            for ti, (tlow, tcounts) in titr:
                # It = slice(tlow, tlow + tcounts)
                fitr = enumerate(zip(freq_mapping[ms][idt]['start_indices'],
                                     freq_mapping[ms][idt]['counts']))
                for fi, (flow, fcounts) in fitr:
                    Inu = slice(flow, flow + fcounts)
                    out_freqs.append(np.mean(freqs[ms][idt][Inu]))
                    ntasks += 1

                    # divide tlow:tlow+tcounts into ipi intervals
                    for t0 in range(tlow, tlow+tcounts, opts.integrations_per_image):
                        tmax = np.minimum(tlow+tcounts, t0 + opts.integrations_per_image)
                        It = slice(t0, tmax)
                        out_times.append(np.mean(utimes[ms][idt][It]))
    
    # remove duplicates
    out_times = np.unique(out_times)
    out_freqs = np.unique(out_freqs)
    n_times = out_times.size
    n_freqs = out_freqs.size

    # if not stacking we only run this to get the number of tasks that will be submitted
    if not opts.stack:
        return None, ntasks, n_times, n_freqs

    # spatial coordinates
    if opts.phase_dir is None:
        out_ra_deg = np.rad2deg(np.unique(out_ra))
        out_dec_deg = np.rad2deg(np.unique(out_dec))
        if out_ra_deg.size > 1 or out_dec_deg.size > 1:
            raise ValueError('phase-dir must be specified when stacking multiple fields')
    else:
        from pfb.utils.astrometry import format_coords
        from astropy import units
        from astropy.coordinates import SkyCoord
        ra_str, dec_str = opts.phase_dir.split(',')
        coord = SkyCoord(ra_str, dec_str, frame='fk5', unit=(units.hourangle, units.deg))
        out_ra_deg = np.array([coord.ra.value])
        out_dec_deg = np.array([coord.dec.value])

    

    n_stokes = len(opts.product)

    cube_dims = (n_stokes, n_freqs, n_times, ny, nx)
    cube_chunks = (n_stokes, 1, time_chunk, spatial_chunk, spatial_chunk)

    mean_dims = (n_stokes, n_freqs, ny, nx)
    mean_chunks = (n_stokes, 1, spatial_chunk, spatial_chunk)

    rms_dims = (n_stokes, n_freqs, n_times)
    rms_chunks = (n_stokes, 1, time_chunk)

    ra_dim = 'RA--SIN'
    dec_dim = 'DEC--SIN'
    if out_freqs.size > 1:
        delta_freq = np.diff(out_freqs).min()
    else:
        delta_freq = 1
    if out_times.size > 1:
        delta_time = np.diff(out_times).min()
    else:
        delta_time = 1

    # construct reference header

    from astropy.wcs import WCS
    from astropy.io import fits
    w = WCS(naxis=5)
    w.wcs.ctype = ['RA---SIN', 'DEC--SIN', 'INTEGRATION', 'FREQ', 'STOKES']
    w.wcs.cdelt = [-cell_deg, cell_deg, delta_time, delta_freq, 1]
    w.wcs.cunit = ['deg', 'deg', 's', 'Hz', '']
    w.wcs.crval = [out_ra_deg[0], out_dec_deg[0], out_times[0], out_freqs[0], 1]
    w.wcs.crpix = [1 + nx//2, 1 + ny//2, 1, 1, 1]
    w.wcs.equinox = 2000.0
    wcs_hdr = w.to_header()

    # the order does seem to matter here,
    # especially when using with StreamingHDU
    hdr = fits.Header()
    hdr['SIMPLE'] = (True, 'conforms to FITS standard')
    hdr['BITPIX'] = (-32, 'array data type')
    hdr['NAXIS'] = (5, 'number of array dimensions')
    data_shape = (nx, ny, n_times, n_freqs, n_stokes)
    for i, size in enumerate(data_shape, 1):
        hdr[f'NAXIS{i}'] = (size, f'length of data axis {i}')

    hdr['EXTEND'] = True
    hdr['BSCALE'] = 1.0
    hdr['BZERO'] = 0.0
    hdr['BUNIT'] = 'Jy/beam'
    hdr['EQUINOX'] = 2000.0
    hdr['BTYPE'] = 'Intensity'
    hdr.update(wcs_hdr)
    hdr['TIMESCAL'] = delta_time
    if opts.obs_label is not None:
        hdr['OBSLABEL'] = opts.obs_label

    # if we don't pass these into stokes2im they get overwritten
    attrs={
            "fits_header": list(dict(hdr).items()),
            "radec_dims": (ra_dim, dec_dim),
            "fits_dims": (
                ("X", ra_dim),
                ("Y", dec_dim),
                ("TIME", "TIME"),
                ("FREQ", "FREQ"),
                ("STOKES", "STOKES")
            )
        }

    # TODO - why do we sometimes need to round here?
    out_ras = out_ra_deg + np.arange(nx//2, -(nx//2), -1) * cell_deg
    out_decs = out_dec_deg + np.arange(-(ny//2), ny//2) * cell_deg
    out_ras = np.round(out_ras, decimals=12)
    out_decs = np.round(out_decs, decimals=12)


    dummy_ds = xr.Dataset(
        data_vars={
            "cube": (
                ("STOKES", "FREQ", "TIME", "Y", "X"),
                da.empty(cube_dims, chunks=cube_chunks, dtype=np.float32)
            ),
            "cube_mean":(
                ("STOKES", "FREQ", "Y", "X"),
                da.empty(mean_dims, chunks=mean_chunks, dtype=np.float32)
            ),
            "rms": (
                ("STOKES", "FREQ", "TIME"),
                da.empty(rms_dims, chunks=rms_chunks, dtype=np.float32)
            ),
            "weight": (
                ("STOKES", "FREQ", "TIME"),
                da.empty(rms_dims, chunks=rms_chunks, dtype=np.float32)
            ),
            "nonzero": (
                ("STOKES", "FREQ", "TIME",),
                da.empty(rms_dims, chunks=rms_chunks, dtype=np.bool_)
            ),
            "psf_maj": (
                    ("STOKES", "FREQ", "TIME",),
                    da.empty(rms_dims, chunks=rms_chunks, dtype=np.float32)
            ),
            "psf_min": (
                    ("STOKES", "FREQ", "TIME",),
                    da.empty(rms_dims, chunks=rms_chunks, dtype=np.float32)
            ),
            "psf_pa": (
                    ("STOKES", "FREQ", "TIME",),
                    da.empty(rms_dims, chunks=rms_chunks, dtype=np.float32)
            ),
            "channel_width": (
                    ("FREQ",),
                    da.empty((n_freqs), chunks=(1), dtype=np.float32)
            ),
        },
        coords={
            "TIME": (("TIME",), out_times),
            "STOKES": (("STOKES",), list(sorted(opts.product))),
            "FREQ": (("FREQ",), out_freqs),
            "X": (("X",), out_ras),
            "Y": (("Y",), out_decs)
        }
        ,
        # Store dicts as tuples as zarr doesn't seem to maintain dict order.
        attrs=attrs
    )

    if opts.beam_model is not None:
        dummy_ds['beam_weight'] = (("STOKES", "FREQ", "TIME", "Y", "X"),
                da.empty(cube_dims, chunks=cube_chunks, dtype=np.float32))

    if opts.psf_out:
        nx_psf = good_size(int(opts.psf_relative_size * nx))
        while nx_psf%2:
            nx_psf = good_size(nx_psf+1)
        ny_psf = good_size(int(opts.psf_relative_size * ny))
        while ny_psf%2:
            ny_psf = good_size(ny_psf+1)
        psf_dims = (n_stokes, n_freqs, n_times, ny_psf, nx_psf)
        if opts.psf_relative_size == 1:
            xpsf = "X"
            ypsf = "Y"
        else:
            xpsf = "X_PSF"
            ypsf = "Y_PSF"
            dummy_ds = dummy_ds.assign_coords(
                {'Y_PSF': ((ypsf,), out_dec_deg + np.arange(-(ny_psf//2), ny_psf//2) * cell_deg),
                'X_PSF': ((xpsf,), out_ra_deg + np.arange(nx_psf//2, -(nx_psf//2), -1) * cell_deg)}
            )
        dummy_ds['psf'] = (("STOKES", "FREQ", "TIME", ypsf, xpsf),
                da.empty(psf_dims, chunks=cube_chunks, dtype=np.float32))
        dummy_ds['psf2'] = (("STOKES", ypsf, xpsf),
                da.empty((n_stokes, ny_psf, nx_psf),
                         chunks=(1, spatial_chunk, spatial_chunk), dtype=np.float32))

    # Write scaffold and metadata to disk.
    dummy_ds.to_zarr(cds_url, mode="w", compute=False)
    return attrs, ntasks, n_times, n_freqs



# flake8: noqa
from contextlib import ExitStack
import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('RESTORE')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.restore)
def restore(**kw):
    '''
    Create fits image cubes from data products (eg. restored images).
    '''
    opts = OmegaConf.create(kw)

    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        opts.nthreads = nthreads//2
        ncpu = ncpu//2

    OmegaConf.set_struct(opts, True)

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool, thread_pool_size
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{opts.log_directory}/restore_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    with ExitStack() as stack:
        if opts.nworkers > 1:
            from pfb import set_client
            client = set_client(opts.nworkers, log, stack=stack, client_log_level=opts.log_level)
        else:
            log.info("Faking client")
            from pfb.utils.dist import fake_client
            client = fake_client()
            names = [0]
        ti = time.time()
        _restore(**opts)

    log.info(f"All done after {time.time() - ti}s")

def _restore(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)


    import numpy as np
    from pfb.utils.naming import xds_from_url, xds_from_list
    from pfb.utils.fits import dds2fits
    from pfb.utils.misc import Gaussian2D, fitcleanbeam, convolve2gaussres
    from ducc0.fft import c2c
    from pfb.utils.restoration import restore_image
    from itertools import cycle
    from daskms.fsspec_store import DaskMSStore
    from pfb.utils.naming import get_opts
    from distributed import wait
    try:
        from distributed import get_client
        client = get_client()
        names = list(client.scheduler_info()['workers'].keys())
        from distributed import as_completed
    except:
        from pfb.utils.dist import fake_client
        client = fake_client()
        names = [0]
        as_completed = lambda x: x

    basename = opts.output_filename
    if opts.fits_output_folder is not None:
        fits_oname = opts.fits_output_folder + '/' + basename.split('/')[-1]
    else:
        fits_oname = basename

    dds_name = f'{basename}_{opts.suffix}.dds'
    dds, dds_list = xds_from_url(dds_name)
    dds_store = DaskMSStore(dds_name)
    if '://' in dds_store.url:
        protocol = dds_store.url.split('://')[0]
    else:
        protocol = 'file'

    # get MFS PSF PARS
    try:
        psfpars_mfs = get_opts(dds_store.url,
                            protocol,
                            name='psfparsn_mfs.pkl')
    except:
        log.error_and_raise("Could not load MFS PSF pamaters. "
                            "Run grid worker with psf=true to remake.",
                            RuntimeError)

    if opts.drop_bands is not None:
        ddso = []
        ddso_list = []
        for ds, dsl in zip(dds, dds_list):
            b = int(ds.bandid)
            if b not in opts.drop_bands:
                ddso.append(ds)
                ddso_list.append(dsl)
        dds = ddso
        dds_list = ddso_list

    timeids = np.unique(np.array([int(ds.timeid) for ds in dds]))
    freqs = [ds.freq_out for ds in dds]
    freqs = np.unique(freqs)
    nband = freqs.size
    ntime = timeids.size
    ncorr = dds[0].corr.size
    cell_asec = dds[0].cell_rad * 180/np.pi * 3600
    log.info(f'Number of output times = {ntime}')
    log.info(f'Number of output bands = {nband}')


    if opts.gausspar is None:
        gaussparf = None
        gaussparf_mfs = None
        log.info("Using native resolution")
    elif opts.gausspar == [0,0,0]:
        # This is just to figure out what resolution to convolve to
        dds = xds_from_list(dds_list, nthreads=opts.nthreads,
                            drop_all_but=['PSFPARSN'])
        emaj = 0.0
        emin = 0.0
        pas = []
        for ds in dds:
            gausspars = ds.PSFPARSN.values
            emaj = np.maximum(emaj, gausspars[:, 0].max())
            emin = np.maximum(emin, gausspars[:, 1].max())
            pas.append(np.mean(gausspars[:, 2]))
        pa = np.mean(np.array(pas))
        log.info(f"Using lowest resolution of ({emaj*cell_asec:.3e} asec, "
              f"{emin*cell_asec:.3e} asec, {pa:.3e} degrees) for restored images")
        # these are n pixel units
        gaussparf_mfs = [emaj, emin, pa]
        gaussparf = gaussparf_mfs * ncorr
    else:
        gaussparf_mfs = list(opts.gausspar)
        emaj = gaussparf_mfs[0]
        emin = gaussparf_mfs[1]
        pa = gaussparf_mfs[2]
        if 'i' in opts.outputs.lower():
            log.info(f"Using specified resolution of ({emaj:.3e} asec, {emin:.3e} asec, "
                f"{pa:.3e} degrees) for restored image")
        # convert to pixel units
        for i in range(2):
            gaussparf_mfs[i] /= cell_asec
        gaussparf = [gaussparf_mfs]*ncorr

    # create restored images
    futrestore = []
    for ds_name in dds_list:
        fut = client.submit(restore_image,
                            ds_name,
                            opts.model_name,
                            opts.residual_name,
                            gaussparf=gaussparf,
                            nthreads=opts.nthreads)
        futrestore.append(fut)

    futures = []
    if 'd' in opts.outputs.lower():
        fut = client.submit(
                        dds2fits,
                        dds_list,
                        'DIRTY',
                        f'{fits_oname}_{opts.suffix}',
                        norm_wsum=True,
                        nthreads=opts.nthreads,
                        do_mfs='d' in opts.outputs,
                        do_cube='D' in opts.outputs,
                        psfpars_mfs=psfpars_mfs)
        futures.append(fut)

    if 'm' in opts.outputs.lower():
        fut = client.submit(
                        dds2fits,
                        dds_list,
                        opts.model_name,
                        f'{fits_oname}_{opts.suffix}',
                        norm_wsum=False,
                        nthreads=opts.nthreads,
                        do_mfs='m' in opts.outputs,
                        do_cube='M' in opts.outputs,
                        psfpars_mfs=psfpars_mfs)
        futures.append(fut)

    if 'r' in opts.outputs.lower():
        fut = client.submit(
                        dds2fits,
                        dds_list,
                        opts.residual_name,
                        f'{fits_oname}_{opts.suffix}',
                        norm_wsum=True,
                        nthreads=opts.nthreads,
                        do_mfs='r' in opts.outputs,
                        do_cube='R' in opts.outputs,
                        psfpars_mfs=psfpars_mfs)
        futures.append(fut)

    if opts.nworkers > 1:
        wait(futrestore)

    if 'i' in opts.outputs.lower():
        fut = client.submit(
                        dds2fits,
                        dds_list,
                        'IMAGE',
                        f'{fits_oname}_{opts.suffix}',
                        norm_wsum=False,
                        nthreads=opts.nthreads,
                        do_mfs='i' in opts.outputs,
                        do_cube='I' in opts.outputs,
                        psfpars_mfs=psfpars_mfs)
        futures.append(fut)

    # TODO(LB) - we may want to add these outputs back in, at least the useful ones
    # if 'f' in opts.outputs:
    #     rhat_mfs = c2c(residual_mfs, forward=True,
    #                    nthreads=opts.nthreads, inorm=0)
    #     rhat_mfs = np.fft.fftshift(rhat_mfs)
    #     save_fits(np.abs(rhat_mfs),
    #               f'{fits_oname}_{opts.suffix}.abs_fft_residual_mfs.fits',
    #               hdr_mfs,
    #               overwrite=opts.overwrite)
    #     save_fits(np.angle(rhat_mfs),
    #               f'{fits_oname}_{opts.suffix}.phase_fft_residual_mfs.fits',
    #               hdr_mfs,
    #               overwrite=opts.overwrite)

    # if 'F' in opts.outputs:
    #     rhat = c2c(residual, axes=(1,2), forward=True,
    #                nthreads=opts.nthreads, inorm=0)
    #     rhat = np.fft.fftshift(rhat, axes=(1,2))
    #     save_fits(np.abs(rhat),
    #               f'{fits_oname}_{opts.suffix}.abs_fft_residual.fits',
    #               hdr,
    #               overwrite=opts.overwrite)
    #     save_fits(np.angle(rhat),
    #               f'{fits_oname}_{opts.suffix}.phase_fft_residual.fits',
    #               hdr,
    #               overwrite=opts.overwrite)

    # if 'c' in opts.outputs:
    #     if GaussPar is None:
    #         raise ValueError("Clean beam in output but no PSF in dds")
    #     cpsf_mfs = Gaussian2D(xx, yy, GaussPar[0], normalise=False)
    #     save_fits(cpsf_mfs,
    #               f'{fits_oname}_{opts.suffix}.cpsf_mfs.fits',
    #               hdr_mfs,
    #               overwrite=opts.overwrite)

    # if 'C' in opts.outputs:
    #     if GaussPars is None:
    #         raise ValueError("Clean beam in output but no PSF in dds")
    #     cpsf = np.zeros(residual.shape, dtype=output_type)
    #     for v in range(opts.nband):
    #         gpar = GaussPars[v]
    #         if not np.isnan(gpar).any():
    #             cpsf[v] = Gaussian2D(xx, yy, gpar, normalise=False)
    #     save_fits(cpsf,
    #               f'{fits_oname}_{opts.suffix}.cpsf.fits',
    #               hdr,
    #               overwrite=opts.overwrite)

    if opts.nworkers > 1:
        wait(futures)

    return
